#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from __future__ import print_function

import os

try:
  import dbm
except ImportError:
  import anydbm

from . import inchi_key
from . import inchi as inchimod



class Config(object):
    database_file = os.path.abspath(os.path.join(os.path.dirname(__file__), "names.db"))



def normalize_inchi( inchi):
    if inchi.startswith("InChI="):
        return inchi[6:]
    else:
        return inchi


def compound_to_database_string( c):
    c['inchi'] = normalize_inchi( c['inchi'])
    c['inchikey'] = inchi_key.key_from_inchi( c['inchi'])
    return "%(inchikey)s %(cid)s ### %(name)s\n" % c


def database_string_to_compound( line):
    a,name = line.split( "###")
    inchi, cid = a.split()
    inchikey = inchi_key.key_from_inchi( inchi)
    return {'inchikey':inchikey.strip(), 'cid':cid.strip(), 'name':name.strip()}


def mydb_to_gdbm( infilename, outfilename):
    import gdbm
    with open(infilename, 'r') as infile:
      base = gdbm.open(outfilename, 'n')
      for line in infile:
          rec = database_string_to_compound(line)
          base[rec['inchikey']] = rec['cid'] + " " + rec['name']


def get_compound_from_database( inchikey, database_file=None):
    #inchi = normalize_inchi( inchi)
    for fname in (database_file, Config.database_file):
        if fname and os.path.exists(fname):
            break
    else:
        raise Exception("Name database not found")
    base = anydbm.open( fname)
    if inchikey in base:
        cid, name = base[ inchikey].split( " ", 1)
        return {'inchikey': inchikey, 'cid': cid, 'name': name}
    else:
        return None


def name_molecule( mol, database_file=None):
    """tries to find name for an OASA molecule in the database,
    it requires InChI generation to work"""
    key = inchimod.generate_inchi_key( mol)
    return get_compound_from_database( key, database_file=database_file)


if __name__ == "__main__":
    import sys
    if len( sys.argv) > 1:
        fname = sys.argv[1]
        if os.path.exists( fname):
            try:
                mydb_to_gdbm( fname, Config.database_file)
            except:
                print("given file must be a text file with one compound per line in format 'InChI CID ### name'")
        else:
            print("you must supply a valid filename to update the database or no argument for a test to run")
    else:
        print(get_compound_from_database( "IJDNQMDRQITEOD-UHFFFAOYSA-N"))



#--------------------------------------------------------------------------
#     This file is part of BKChem - a chemical drawing program
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------
#
#
#
#--------------------------------------------------------------------------

"""Module containing miscelanous functions used in BKChem that don't
fit anywhere else. Does not contain any objects.
"""

import re
import sys
import math
import time
import types
import string
import base64

from warnings import warn



def myisstr(obj):
  if sys.version_info[0] > 2:
    return isinstance(obj, str)
  else:
    return isinstance(obj, basestring)


def intersection( a, b):
  "returns intersection of 2 lists"
  ret = []
  for i in a:
    if i in b:
      ret.append( i)
  return ret


def difference( a,b):
  "returns difference of 2 lists ( a-b)"
  ret = list( a)  # needed for type conversion of tuple for instance
  for i in b:
    if i in ret:
      ret.remove( i)
  return ret


def signum( a):
  if a == 0:
    return 0
  elif a < 0:
    return -1
  else:
    return 1


def filter_unique( items):
  #t1 = time.time()
  ret = []
  for item in items:
    if item not in ret:
      ret.append( item)
  #print time.time() - t1
  return ret


#def sub_all_numbers( text):
#  return re.sub( "\d+", '<sub>\g<0><\\sub>', text)

def reverse_molecule_formula( formula):
  pass
  

def normalize_coords( coords):
  x1, y1, x2, y2 = coords
  if x2 < x1:
    x2, x1 = x1, x2
  if y2 < y1:
    y2, y1 = y1, y2
  return (x1, y1, x2, y2)


def list_difference( list):
  """return a list of differences between list members,
  the list is by 1 shorter than the original"""
  ret = []
  for i in range( len( list)-1):
    ret.append( list[i]-list[i+1])
  return ret


def split_number_and_unit( txt):
  try:
    v = float( txt)
    return v, ''
  except:
    pass
  cutter = re.compile( "([+-]?\d*\.?\d*)\s*([a-zA-Z]*)")
  if txt:
    a = cutter.match( txt)
    if a and a.group(1):
      return float( a.group(1)), a.group(2)
  return None, None


def lazy_apply( function, arguments):
  """similar to apply but returns a callable (lambda) that performs the apply when called."""
  return lambda: apply( function, arguments)




def extend_bbox( bbox, pixels=1):
  minx = min( (bbox[0], bbox[2]))
  maxx = max( (bbox[0], bbox[2]))
  miny = min( (bbox[1], bbox[3]))
  maxy = max( (bbox[1], bbox[3]))
  return minx-pixels, miny-pixels, maxx+pixels, maxy+pixels


def smallest_common_bbox( bboxes):
  _x0, _y0, _x1, _y1 = None, None, None, None
  for (x0, y0, x1, y1) in bboxes:
    minx = min( x0, x1)
    maxx = max( x0, x1)
    miny = min( y0, y1)
    maxy = max( y0, y1)
    if not _x0 or minx < _x0:
      _x0 = minx
    if not _x1 or maxx > _x1:
      _x1 = maxx
    if not _y0 or miny < _y0:
      _y0 = miny
    if not _y1 or maxy > _y1:
      _y1 = maxy
  return _x1, _y1, _x0, _y0

      
    



def has_one_value_only( iterable):
  if not iterable:
    return 0
  a = iterable[0]
  for i in iterable:
    if a != i:
      return 0
  return 1



def plural_or_singular( iterable):
  """useful for string construction such as 'you have %d apple%s' % (len(apples), plural_or_singular( apples)"""
  if len( iterable) == 1:
    return ''
  else:
    return 's'


def len_and_ending( iterable):
  return (len( iterable), plural_or_singular( iterable))


def reverse( iterable):
  for i in range( len( iterable)-1, -1, -1):
    yield iterable[i]



# some helper, higher order functions

map_functions = lambda funcs, value: zip( apply, funcs, len(funcs)*[value])

something_true = lambda vals: len( filter( None, vals))

some_apply = lambda func, vals: something_true( map( func, vals))



# some results

isinstance_of_one = lambda obj, parents: some_apply( lambda x: isinstance( obj, x), parents)


  

def gen_variations(items, n):
  if n==0:
    yield []
  else:
    for i in range(len(items)-n+1):
      for v in gen_variations(items[i+1:],n-1):
        yield [items[i]]+v


def gen_combinations(items, n):
  if n==0:
    yield []
  else:
    for i in range(len(items)):
      for v in gen_combinations([x for x in items if x!=items[i]], n-1):
        yield [items[i]]+v


def x_over_y( x, y):
  return factorial( x) / factorial( y) / factorial( x-y)


def factorial( x):
  ret = 1
  for i in range( 2, x+1):
    ret *= i
  return ret


def gen_variations_and_one( items, length):
  for i in items:
    vars = gen_variations( [x for x in items if x!=i], length-1)
    for vs in vars:
      yield vs + [i]




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

import sys
import copy
import math
import cairo

from . import misc
from . import geometry
from . import transform3d



class cairo_out(object):
  """Draw OASA molecules using cairo.

  Cairo supports different 'surfaces' which represent different file formats.
  This object implements PNG file drawing, but should be general enough
  to work with other formats, provided modified version of create_surface and
  write_surface methods are provided when this class is subclassed.

  Usage:

  c = cairo_out( scaling=2.0, margin=10, font_size=20, bond_width=6)
  c.show_hydrogens_on_hetero = True
  c.mol_to_cairo( mol, 'outfile.png') # mol is oasa molecule

  # attributes can be set in constructor or afterwards as normal attributes
  # default options are set and described in the cairo_out.default_options dictionary
  """

  _temp_margin = 200

  atom_colors_minimal = {'O': (1,0,0),
                         'N': (0,0,1),
                         'S': (.5,.5,0),
                         'Cl': (0,0.8,0),
                         'Br': (.5,0,0),
                         }

  ## taken from: http://bodr.svn.sourceforge.net/viewvc/*checkout*/bodr/trunk/bodr/elements/elements.xml?revision=34&content-type=text%2Fplain
  ## C and H modified to be black
  ## F,S,Cl modified to a darker color (suggested by Toon Verstraelen)
  atom_colors_full = {
    'H': (0.00,0.00,0.00), 'He': (0.85,1.00,1.00),
    'Li': (0.80,0.50,1.00), 'Be': (0.76,1.00,0.00), 'B': (1.00,0.71,0.71),
    'C': (0.0,0.0,0.0), 'N': (0.05,0.05,1.00), 'O': (1.00,0.05,0.05),
    'F': (0.30,0.70,0.70), 'Ne': (0.70,0.89,0.96), 'Na': (0.67,0.36,0.95),
    'Mg': (0.54,1.00,0.00), 'Al': (0.75,0.65,0.65), 'Si': (0.50,0.60,0.60),
    'P': (1.00,0.50,0.00), 'S': (0.80,0.80,0.20), 'Cl': (0.00,0.80,0.00),
    'Ar': (0.50,0.82,0.89), 'K': (0.56,0.25,0.83), 'Ca': (0.24,1.00,0.00),
    'Sc': (0.90,0.90,0.90), 'Ti': (0.75,0.76,0.78), 'V': (0.65,0.65,0.67),
    'Cr': (0.54,0.60,0.78), 'Mn': (0.61,0.48,0.78), 'Fe': (0.50,0.48,0.78),
    'Co': (0.44,0.48,0.78), 'Ni': (0.36,0.48,0.76), 'Cu': (1.00,0.48,0.38),
    'Zn': (0.49,0.50,0.69), 'Ga': (0.76,0.56,0.56), 'Ge': (0.40,0.56,0.56),
    'As': (0.74,0.50,0.89), 'Se': (1.00,0.63,0.00), 'Br': (0.65,0.16,0.16),
    'Kr': (0.36,0.72,0.82), 'Rb': (0.44,0.18,0.69), 'Sr': (0.00,1.00,0.00),
    'Y': (0.58,1.00,1.00), 'Zr': (0.58,0.88,0.88), 'Nb': (0.45,0.76,0.79),
    'Mo': (0.33,0.71,0.71), 'Tc': (0.23,0.62,0.62), 'Ru': (0.14,0.56,0.56),
    'Rh': (0.04,0.49,0.55), 'Pd': (0.00,0.41,0.52), 'Ag': (0.88,0.88,1.00),
    'Cd': (1.00,0.85,0.56), 'In': (0.65,0.46,0.45), 'Sn': (0.40,0.50,0.50),
    'Sb': (0.62,0.39,0.71), 'Te': (0.83,0.48,0.00), 'I': (0.58,0.00,0.58),
    'Xe': (0.26,0.62,0.69), 'Cs': (0.34,0.09,0.56), 'Ba': (0.00,0.79,0.00),
    'La': (0.44,0.83,1.00), 'Ce': (1.00,1.00,0.78), 'Pr': (0.85,1.00,0.78),
    'Nd': (0.78,1.00,0.78), 'Pm': (0.64,1.00,0.78), 'Sm': (0.56,1.00,0.78),
    'Eu': (0.38,1.00,0.78), 'Gd': (0.27,1.00,0.78), 'Tb': (0.19,1.00,0.78),
    'Dy': (0.12,1.00,0.78), 'Ho': (0.00,1.00,0.61), 'Er': (0.00,0.90,0.46),
    'Tm': (0.00,0.83,0.32), 'Yb': (0.00,0.75,0.22), 'Lu': (0.00,0.67,0.14),
    'Hf': (0.30,0.76,1.00), 'Ta': (0.30,0.65,1.00), 'W': (0.13,0.58,0.84),
    'Re': (0.15,0.49,0.67), 'Os': (0.15,0.40,0.59), 'Ir': (0.09,0.33,0.53),
    'Pt': (0.96,0.93,0.82), 'Au': (0.80,0.82,0.12), 'Hg': (0.71,0.71,0.76),
    'Tl': (0.65,0.33,0.30), 'Pb': (0.34,0.35,0.38), 'Bi': (0.62,0.31,0.71),
    'Po': (0.67,0.36,0.00), 'At': (0.46,0.31,0.27), 'Rn': (0.26,0.51,0.59),
    'Fr': (0.26,0.00,0.40), 'Ra': (0.00,0.49,0.00), 'Ac': (0.44,0.67,0.98),
    'Th': (0.00,0.73,1.00), 'Pa': (0.00,0.63,1.00), 'U': (0.00,0.56,1.00),
    'Np': (0.00,0.50,1.00), 'Pu': (0.00,0.42,1.00), 'Am': (0.33,0.36,0.95),
    'Cm': (0.47,0.36,0.89), 'Bk': (0.54,0.31,0.89), 'Cf': (0.63,0.21,0.83),
    'Es': (0.70,0.12,0.83), 'Fm': (0.70,0.12,0.73), 'Md': (0.70,0.05,0.65),
    'No': (0.74,0.05,0.53), 'Lr': (0.78,0.00,0.40), 'Rf': (0.80,0.00,0.35),
    'Db': (0.82,0.00,0.31), 'Sg': (0.85,0.00,0.27), 'Bh': (0.88,0.00,0.22),
    'Hs': (0.90,0.00,0.18), 'Mt': (0.91,0.00,0.15), 'Ds': (0.92,0.00,0.14),
    'Rg': (0.93,0.00,0.13),
    }

  # all the following values are settable using the constructor, e.g.
  # cairo_out( margin=20, bond_width=3.0)
  # all metrics is scaled properly, the values corespond to pixels only
  # when scaling is 1.0
  default_options = {
    'scaling': 1.0,
    # should atom coordinates be rounded to whole pixels before rendering?
    # This improves image sharpness but might slightly change the geometry
    'align_coords': True,
    # the following also draws hydrogens on shown carbons
    'show_hydrogens_on_hetero': False,
    'show_carbon_symbol': False,
    'margin': 15,
    'line_width': 1.0,
    # how far second bond is drawn
    'bond_width': 6.0,
    'wedge_width': 6.0,
    'font_name': "Arial",
    'font_size': 16,
    # background color in RGBA
    'background_color': (1,1,1,1),
    'color_atoms': True,
    'color_bonds': True,
    'space_around_atom': 2,
    # you can choose between atom_colors_full, atom_colors_minimal
    # or provide a custom dictionary
    'atom_colors': atom_colors_full,
    # proportion between subscript and normal letters size
    'subscript_size_ratio': 0.8,
    # how much to shorten second line of double and triple bonds (0-0.5)
    'bond_second_line_shortening': 0.15,
    # the following two are just for playing
    # - without antialiasing the output is ugly
    'antialias_text': True,
    'antialias_drawing': True,
    # this will only change appearance of overlapping text
    'add_background_to_text': False,
    # should text be converted to curves?
    'text_to_curves': False,
    }


  def __init__( self, **kw):
    for k, v in self.__class__.default_options.items():
      setattr( self, k, v)
    # list of paths that contribute to the bounding box (probably no edges)
    self._vertex_to_bbox = {} # vertex-to-bbox mapping
    self._bboxes = [] # for overall bbox calcualtion
    for k,v in kw.items():
      if k in self.__class__.default_options:
        setattr( self, k, v)
      else:
        raise Exception( "unknown attribute '%s' passed to constructor" % k)


  def draw_mol( self, mol):
    if not self.surface:
      raise Exception( "You must initialize cairo surface before drawing, use 'create_surface' to do it.")
    self.molecule = mol
    for v in mol.vertices:
      self._draw_vertex( v)
    for e in copy.copy( mol.edges):
      self._draw_edge( e)


  def create_surface( self, w, h, format):
    """currently implements PNG writting, but might be overriden to write other types;
    w and h are minimal estimated width and height"""
    # trick - we use bigger surface and then copy from it to a new surface and crop
    if format == "png":
      self.surface = cairo.ImageSurface( cairo.FORMAT_ARGB32, w, h)
    elif format == "pdf":
      self.surface = cairo.PDFSurface( self.filename, w, h)
    elif format == "svg":
      self.surface = cairo.SVGSurface( self.filename, w, h)
    else:
      raise Exception( "unknown format '%s'" % format)


  def init_surface( self):
    """make all necessary operations to prepare a surface for drawing:
    set antialiasing as requested, create background, etc."""
    pass


  def create_dummy_surface( self, w, h):
    self.surface = cairo.ImageSurface( cairo.FORMAT_A1, w, h)


  def write_surface( self):
    """finishes the surface and write it to the file if necessary"""
    self.context.show_page()
    if isinstance( self.surface, cairo.ImageSurface):
      self.surface.write_to_png( self.filename)
    self.surface.finish()


  def mols_to_cairo( self, mols, filename, format="png"):
    x1, y1, x2, y2 = None, None, None, None
    for mol in mols:
      for v in mol.vertices:
        v.y = -v.y # flip coords - molfiles have them the other way around
        if self.align_coords:
          v.x = self._round( v.x)
          v.y = self._round( v.y)
        if x1 is None or x1 > v.x:
          x1 = v.x
        if x2 is None or x2 < v.x:
          x2 = v.x
        if y1 is None or y1 > v.y:
          y1 = v.y
        if y2 is None or y2 < v.y:
          y2 = v.y
    w = int( x2 - x1)
    h = int( y2 - y1)
    self._bboxes.append( (x1,y1,x2,y2))

    # dummy surface to get complete size of the drawing
    # Because it is not possible to calculate the bounding box of a drawing before its drawn (mainly
    # because we don't know the size of text items), this object internally draws to a surface with
    # large margins to get the bbox
    _w = int( self.scaling*w+2*self.scaling*self._temp_margin)
    _h = int( self.scaling*h+2*self.scaling*self._temp_margin)
    self.create_dummy_surface( _w, _h)
    self.context = cairo.Context( self.surface)
    [self.draw_mol( mol) for mol in mols]
    x1, y1, x2, y2 = self._get_bbox()
    x1, y1 = self.context.user_to_device( x1, y1)
    x2, y2 = self.context.user_to_device( x2, y2)
    width = int( self.scaling*(x2-x1) + 2*self.margin*self.scaling)
    height = int( self.scaling*(y2-y1) + 2*self.margin*self.scaling)

    # now paint for real
    self.filename = filename
    self.create_surface( width, height, format)
    self.context = cairo.Context( self.surface)
    if not self.antialias_drawing:
      self.context.set_antialias( cairo.ANTIALIAS_NONE)
    if not self.antialias_text:
      options = self.context.get_font_options()
      options.set_antialias( cairo.ANTIALIAS_NONE)
      self.context.set_font_options( options)
    self.context.translate( round( -x1*self.scaling+self.scaling*self.margin), round( -y1*self.scaling+self.scaling*self.margin))
    self.context.scale( self.scaling, self.scaling)
    self.context.rectangle( x1, y1, w, h)
    self._set_source_color( self.background_color)
    self.context.paint()
    self.context.new_path()
    self.context.set_source_rgb( 0, 0, 0)
    [self.draw_mol( mol) for mol in mols]
    # write the content to the file
    self.write_surface()
    # flip y coordinates back
    for v in mol.vertices:
      v.y = -v.y


  def mol_to_cairo( self, mol, filename, format="png"):
    """This is a convenience method kept for backward compatibility,
    it just calls mols_to_cairo internally"""
    return self.mols_to_cairo( [mol], filename, format=format)


  def _round( self, x):
    if self.line_width % 2:
      return round( x) + 0.5
    return round( x)


  def _draw_edge( self, e):
    def draw_plain_or_colored_line( _start, _end, second=False):
      """second means if this is not the main line, drawing might be different"""
      if not has_shown_vertex or not self.color_bonds:
        if not second:
          self._draw_line( _start, _end, line_width=self.line_width, capstyle=cairo.LINE_CAP_ROUND)
        else:
          self._draw_line( _start, _end, line_width=self.line_width, capstyle=cairo.LINE_CAP_BUTT)
      else:
        self._draw_colored_line( _start, _end, line_width=self.line_width, start_color=color1, end_color=color2)


    def draw_plain_or_colored_wedge( _start, _end):
      x1, y1 = _start
      x2, y2 = _end
      x, y, x0, y0 = geometry.find_parallel( x1, y1, x2, y2, self.wedge_width/2.0)
      xa, ya, xb, yb = geometry.find_parallel( x1, y1, x2, y2, self.line_width/2.0)
      # no coloring now
      if not has_shown_vertex or not self.color_bonds:
        self._create_cairo_path( [(xa, ya), (x0, y0), (2*x2-x0, 2*y2-y0), (2*x1-xa, 2*y1-ya)], closed=True)
        self.context.set_source_rgb( 0,0,0)
        self.context.fill()
      else:
        # ratio 0.4 looks better than 0.5 because the area difference
        # is percieved more than length difference
        ratio = 0.4
        xm1 = ratio*xa + (1-ratio)*x0
        ym1 = ratio*ya + (1-ratio)*y0
        xm2 = (1-ratio)*(2*x2-x0) + ratio*(2*x1-xa)
        ym2 = (1-ratio)*(2*y2-y0) + ratio*(2*y1-ya)
        self.context.set_source_rgb( *color1)
        self._create_cairo_path( [(xa,ya), (xm1,ym1), (xm2,ym2), (2*x1-xa, 2*y1-ya)], closed=True)
        self.context.fill()
        self.context.set_source_rgb( *color2)
        self._create_cairo_path( [(xm1,ym1), (x0, y0), (2*x2-x0, 2*y2-y0), (xm2,ym2)], closed=True)
        self.context.fill()


    def draw_plain_or_colored_hatch( _start, _end):
      x1, y1 = _start
      x2, y2 = _end
      # no coloring now
      x, y, x0, y0 = geometry.find_parallel( x1, y1, x2, y2, self.wedge_width/2.0)
      xa, ya, xb, yb = geometry.find_parallel( x1, y1, x2, y2, self.line_width/2.0)
      d = math.sqrt( (x1-x2)**2 + (y1-y2)**2) # length of the bond
      if d == 0:
        return  # to prevent division by zero
      dx1 = (x0 - xa)/d
      dy1 = (y0 - ya)/d
      dx2 = (2*x2 -x0 -2*x1 +xa)/d
      dy2 = (2*y2 -y0 -2*y1 +ya)/d
      # we have to decide if the first line should be at the position of the first atom
      draw_start = 1  # is index not boolean
      if not v1 in self._vertex_to_bbox and v1.occupied_valency > 1:
        draw_start = 1
      draw_end = 1    # is added to index not boolean
      if not v2 in self._vertex_to_bbox and v2.occupied_valency > 1:
        draw_end = 0
      # adjust the step length
      step_size = 2*(self.line_width)
      ns = round( d / step_size) or 1
      step_size = d / ns
      # now we finally draw
      self.context.set_line_cap( cairo.LINE_CAP_BUTT)
      self.context.set_source_rgb( *color1)
      middle = 0.5 * (draw_start + int( round( d/ step_size)) + draw_end - 2)
      for i in range( draw_start, int( round( d/ step_size)) +draw_end):
        coords = [xa+dx1*i*step_size, ya+dy1*i*step_size, 2*x1-xa+dx2*i*step_size, 2*y1-ya+dy2*i*step_size]
        if coords[0] == coords[2] and coords[1] == coords[3]:
          if (dx1+dx2) > (dy1+dy2):
            coords[0] += 1
          else:
            coords[1] += 1
        self._create_cairo_path( [coords[:2],coords[2:]])
        if i >= middle:
          self.context.stroke()
          self.context.set_source_rgb( *color2)
      self.context.stroke()

    # code itself
    # at first detect the need to make 3D adjustments
    self._transform = transform3d.transform3d()
    self._invtransform = transform3d.transform3d()
    transform = None
    if e.order > 1:
      atom1,atom2 = e.vertices
      for n in atom1.neighbors + atom2.neighbors:
        # e.atom1 and e.atom2 are in this list as well
        if n.z != 0:
          # engage 3d transform prior to detection of where to draw
          transform = self._get_3dtransform_for_drawing( e)
          #transform = None
          break
      if transform:
        for n in atom1.neighbors + atom2.neighbors:
          n.coords = transform.transform_xyz( *n.coords)
        self._transform = transform
        self._invtransform = transform.get_inverse()
    # // end of 3D adjustments
    # now the code itself
    coords = self._where_to_draw_from_and_to( e)
    if coords:
      start = coords[:2]
      end = coords[2:]
      v1, v2 = e.vertices
      if self.color_bonds:
        color1 = self.atom_colors.get( v1.symbol, (0,0,0))
        color2 = self.atom_colors.get( v2.symbol, (0,0,0))
      else:
        color1 = color2 = (0,0,0)
      has_shown_vertex = bool( [1 for _v in e.vertices if _v in self._vertex_to_bbox])

      if e.order == 1:
        if e.type == 'w':
          draw_plain_or_colored_wedge( start, end)
        elif e.type == 'h':
          draw_plain_or_colored_hatch( start, end)
        else:
          draw_plain_or_colored_line( start, end)

      if e.order == 2:
        side = 0
        # find how to center the bonds
        # rings have higher priority in setting the positioning
        in_ring = False
        for ring in self.molecule.get_smallest_independent_cycles_dangerous_and_cached():
          double_bonds = len( [b for b in self.molecule.vertex_subgraph_to_edge_subgraph(ring) if b.order == 2])
          if v1 in ring and v2 in ring:
            in_ring = True
            side += double_bonds * sum(geometry.on_which_side_is_point(start + end, (a.x, a.y))
                                           for a in ring
                                               if a != v1 and a != v2)
        # if rings did not decide, use the other neigbors
        if not side:
          for v in v1.neighbors + v2.neighbors:
            if v != v1 and v!= v2:
              side += geometry.on_which_side_is_point( start+end, (v.x, v.y))
        # if neighbors did not decide either
        if not side and (in_ring or not has_shown_vertex):
          if in_ring:
            # we don't want centered bonds inside rings
            side = 1 # select arbitrary value
          else:
            # bond between two unshown atoms - we want to center them only in some cases
            if len( v1.neighbors) == 1 and len( v2.neighbors) == 1:
              # both atoms have only one neighbor
              side = 0
            elif len( v1.neighbors) < 3 and len( v2.neighbors) < 3:
              # try to figure out which side is more towards the center of the molecule
              side = sum(geometry.on_which_side_is_point(start + end, (a.x,a.y))
                             for a in self.molecule.vertices
                                 if a != v1 and a != v2)
              if not side:
                side = 1 # we choose arbitrary value, we don't want centering
        if side:
          draw_plain_or_colored_line( start, end)
          x1, y1, x2, y2 = geometry.find_parallel( start[0], start[1], end[0], end[1], self.bond_width*misc.signum( side))
          # shorten the second line
          length = geometry.point_distance( x1,y1,x2,y2)
          if v2 not in self._vertex_to_bbox:
            x2, y2 = geometry.elongate_line( x1, y1, x2, y2, -self.bond_second_line_shortening*length)
          if v1 not in self._vertex_to_bbox:
            x1, y1 = geometry.elongate_line( x2, y2, x1, y1, -self.bond_second_line_shortening*length)
          draw_plain_or_colored_line( (x1, y1), (x2, y2), second=True)
        else:
          for i in (1,-1):
            x1, y1, x2, y2 = geometry.find_parallel( start[0], start[1], end[0], end[1], i*self.bond_width*0.5)
            draw_plain_or_colored_line( (x1, y1), (x2, y2))

      elif e.order == 3:
        draw_plain_or_colored_line( start, end)
        for i in (1,-1):
          x1, y1, x2, y2 = geometry.find_parallel( start[0], start[1], end[0], end[1], i*self.bond_width*0.7)
          draw_plain_or_colored_line( (x1, y1), (x2, y2), second=True)

    if transform:
      # if transform was used, we need to transform back
      for n in atom1.neighbors + atom2.neighbors:
        n.coords = self._invtransform.transform_xyz( *n.coords)


  def _where_to_draw_from_and_to( self, b):
    def fix_bbox( a):
      x, y = a.x, a.y
      data = self._vertex_to_bbox.get( a, None)
      if data:
        (ox, oy), bbox = data
        dx = x - ox
        dy = y - oy
        bbox = [bbox[0]+dx,bbox[1]+dy,bbox[2]+dx,bbox[3]+dy]
        return bbox
      return None
    # at first check if the bboxes are not overlapping
    atom1, atom2 = b.vertices
    x1, y1 = atom1.x, atom1.y
    x2, y2 = atom2.x, atom2.y
    bbox1 = fix_bbox( atom1)
    bbox2 = fix_bbox( atom2)
    if bbox1 and bbox2 and geometry.do_rectangles_intersect( bbox1, bbox2):
      return None
    # then we continue with computation
    if bbox1:
      x1, y1 = geometry.intersection_of_line_and_rect( (x1,y1,x2,y2), bbox1, round_edges=0)
    if bbox2:
      x2, y2 = geometry.intersection_of_line_and_rect( (x1,y1,x2,y2), bbox2, round_edges=0)
    if geometry.point_distance( x1, y1, x2, y2) <= 1.0:
      return None
    else:
      return (x1, y1, x2, y2)


  def _is_there_place( self, atom, x, y):
    x1, y1 = atom.x, atom.y
    angle1 = geometry.clockwise_angle_from_east( x-x1, y-y1)
    for n in atom.neighbors:
      angle = geometry.clockwise_angle_from_east( n.x-x1, n.y-y1)
      if abs( angle - angle1) < 0.3:
        return False
    return True


  def _find_place_around_atom( self, atom):
    x, y = atom.x, atom.y
    coords = [(a.x,a.y) for a in atom.neighbors]
    # now we can compare the angles
    angles = [geometry.clockwise_angle_from_east( x1-x, y1-y) for x1,y1 in coords]
    angles.append( 2*math.pi + min( angles))
    angles = sorted(angles, reverse=True)
    diffs = misc.list_difference( angles)
    i = diffs.index( max( diffs))
    angle = (angles[i] +angles[i+1]) / 2
    return angle


  def _draw_vertex( self, v):
    pos = sum( [(a.x < v.x) and -1 or 1 for a in v.neighbors if abs(a.x-v.x)>0.2])
    if 'show_symbol' in v.properties_:
      show_symbol = v.properties_['show_symbol']
    else:
      show_symbol = (v.symbol != "C" or v.degree == 0 or self.show_carbon_symbol)
    if show_symbol:
      x = v.x
      y = v.y
      text = v.symbol
      # RADICAL
      if v.multiplicity > 1:
        text += "."*(v.multiplicity-1)
      hs = ""
      if self.show_hydrogens_on_hetero or v.properties_.get( 'show_hydrogens', False):
        if v.get_hydrogen_count() == 1:
          hs = "H"
        elif v.get_hydrogen_count() > 1:
          hs = "H<sub>%d</sub>" % v.get_hydrogen_count()
      if not hs:
        pos = -1
      if pos <= 0:
        text += hs
      else:
        text = hs + text
      # charge
      charge = ""
      if v.charge == 1:
        charge = "<sup>+</sup>"
      elif v.charge == -1:
        charge = "<sup>&#x2212;</sup>"
      elif v.charge > 1:
        charge = "<sup>%d+</sup>" % v.charge
      elif v.charge < -1:
        charge = "<sup>%d&#x2212;</sup>" % abs( v.charge)
      if charge:
        if self._is_there_place( v, v.x+3, v.y-2) or v.charge < 0:
          # we place negative charge regardless of available place
          # otherwise minus might be mistaken for a bond
          text += charge
          charge = ""

      # coloring
      if self.color_atoms:
        color = self.atom_colors.get( v.symbol, (0,0,0))
      else:
        color = (0,0,0)
      center_letter = pos <= 0 and 'first' or 'last'
      bbox = self._draw_text( (x,y), text, center_letter=center_letter, color=color)
      self._vertex_to_bbox[v] = ((x,y), geometry.expand_rectangle( bbox, self.space_around_atom))

      # sometimes charge is done here, if it wasn't done before
      if charge:
        assert v.charge > 0
        # if charge was not dealt with we change its appearance from 2+ to ++
        charge = v.charge * "+"
        if self._is_there_place( v, v.x, v.y-10):
          angle = 1.5*math.pi
        elif self._is_there_place( v, v.x, v.y+10):
          angle = 0.5*math.pi
        else:
          angle = self._find_place_around_atom( v)
        self.context.set_font_size( self.subscript_size_ratio * self.font_size)
        xbearing, ybearing, width, height, x_advance, y_advance = self.context.text_extents( charge)
        x0 = v.x + 40*math.cos( angle)
        y0 = v.y + 40*math.sin( angle)
        line = (v.x,v.y,x0,y0)
        charge_bbox = [x0-0.5*width,y0-0.5*height,x0+0.5*width,y0+0.5*height]
        x1, y1 = geometry.intersection_of_line_and_rect( line, bbox, round_edges=0)
        x2, y2 = geometry.intersection_of_line_and_rect( line, charge_bbox, round_edges=0)
        x2, y2 = geometry.elongate_line( x1,y1,x2,y2, -self.space_around_atom)
        x = x0 + x1 - x2
        y = y0 + y1 - y2
        # draw
        self.context.set_source_rgb( *color)
        self.context.move_to( round( x-xbearing-0.5*width), round( y+0.5*height))
        if self.text_to_curves:
          self.context.text_path( charge)
          self.context.fill()
        else:
          self.context.show_text( charge)


  def _get_3dtransform_for_drawing( self, b):
    """this is a helper method that returns a transform3d which rotates
    a bond and its neighbors to coincide with the x-axis and rotates neighbors to be in (x,y)
    plane."""
    atom1, atom2 = b.vertices
    x1,y1,z1 = atom1.coords
    x2,y2,z2 = atom2.coords
    t = geometry.create_transformation_to_coincide_point_with_z_axis( [x1,y1,z1],[x2,y2,z2])
    x,y,z = t.transform_xyz( x2,y2,z2)
    # now rotate to make the plane of neighbor atoms coincide with x,y plane
    angs = []
    for n in atom1.neighbors + atom2.neighbors:
      if n is not atom1 and n is not atom2:
        nx,ny,nz = t.transform_xyz( *n.coords)
        ang = math.atan2( ny, nx)
        if ang < -0.00001:
          ang += math.pi
        angs.append( ang)
    ang = sum( angs) / len( angs)
    t.set_rotation_z( ang + math.pi/2.0)
    t.set_rotation_y( math.pi/2.0)
    return t


  ## ------------------------------ lowlevel drawing methods ------------------------------
  def _draw_colored_line( self, start, end, line_width=1, capstyle=cairo.LINE_CAP_BUTT,
                          start_color=(0,0,0), end_color=(0,0,0)):
    x1,y1 = start
    x2,y2 = end
    length = geometry.point_distance( x1,y1,x2,y2)
    xn2,yn2 = geometry.elongate_line( x1,y1,x2,y2, -0.5*length)
    line1 = [(x1,y1),(xn2,yn2)]
    xn1,yn1 = geometry.elongate_line( x2,y2,x1,y1, -0.5*length)
    line2 = [(xn1,yn1),(x2,y2)]
    self.context.set_line_cap( cairo.LINE_CAP_BUTT) # this is forced here
    self.context.set_line_width( line_width)
    for line,color in zip( [line1,line2], [start_color,end_color]):
      self.context.set_source_rgb( *color)
      self._create_cairo_path( line, closed=False)
      self.context.stroke()


  def _draw_line( self, start, end, line_width=1, capstyle=cairo.LINE_CAP_BUTT, color=(0,0,0)):
    self.context.set_source_rgb( *color)
    # cap style
    self.context.set_line_cap( capstyle)
    # line width
    self.context.set_line_width( line_width)
    # the path itself
    cs = [start, end]
    self._create_cairo_path( cs, closed=False)
    # stroke it
    self.context.stroke()


  def _draw_text( self, xy, text, font_name=None, font_size=None, center_letter=None,
                  color=(0,0,0)):
    import xml.sax
    class text_chunk(object):
      def __init__( self, text, attrs=None):
        self.text = text
        self.attrs = attrs or set()

    class FtextHandler ( xml.sax.ContentHandler):
      def __init__( self):
        xml.sax.ContentHandler.__init__( self)
        self._above = []
        self.chunks = []
        self._text = ""
      def startElement( self, name, attrs):
        self._closeCurrentText()
        self._above.append( name)
      def endElement( self, name):
        self._closeCurrentText()
        self._above.pop( -1)
      def _closeCurrentText( self):
        if self._text:
          self.chunks.append( text_chunk( self._text, attrs = set( self._above)))
          self._text = ""
      def characters( self, data):
        self._text += data

    # parse the text for markup
    handler = FtextHandler()
    try:
      xml.sax.parseString( "<x>%s</x>" % text, handler)
    except:
      chunks = [text_chunk( text)]
    else:
      chunks = handler.chunks

    if not font_name:
      font_name = self.font_name
    if not font_size:
      font_size = self.font_size

    # font properties
    self.context.select_font_face( font_name)
    self.context.set_font_size( font_size)
    asc, desc, letter_height, _a, _b = self.context.font_extents()
    x, y = xy
    if center_letter == 'first':
      if "sup" in chunks[0].attrs or 'sub' in chunks[0].attrs:
        self.context.set_font_size( int( font_size * self.subscript_size_ratio))
      else:
        self.context.set_font_size( font_size)
      xbearing, ybearing, width, height, x_advance, y_advance = self.context.text_extents( chunks[0].text[0])
      x -= 0.5*x_advance
      y += 0.5*height
    elif center_letter == 'last':
      # this is more complicated - we must do a dry run of the text layout
      _dx = 0
      for i,chunk in enumerate( chunks):
        if "sup" in chunk.attrs or 'sub' in chunk.attrs:
          self.context.set_font_size( int( font_size * self.subscript_size_ratio))
        else:
          self.context.set_font_size( font_size)
        xbearing, ybearing, width, height, x_advance, y_advance = self.context.text_extents( chunk.text)
        _dx += x_advance
      # last letter
      xbearing, ybearing, width, height, x_advance, y_advance = self.context.text_extents( chunk.text[-1])
      x -= _dx - 0.5*x_advance
      y += 0.5*height

    self.context.new_path()
    x1 = round( x)
    bbox = None
    for chunk in chunks:
      y1 = round( y)
      if "sup" in chunk.attrs:
        y1 -= asc / 2
        self.context.set_font_size( int( font_size * self.subscript_size_ratio))
      elif "sub" in chunk.attrs:
        y1 += asc / 2
        self.context.set_font_size( int( font_size * self.subscript_size_ratio))
      else:
        self.context.set_font_size( font_size)
      xbearing, ybearing, width, height, x_advance, y_advance = self.context.text_extents( chunk.text)
      # background
      if self.add_background_to_text:
        self.context.rectangle( x1+xbearing, y1+ybearing, width, height)
        self._set_source_color( self.background_color)
        self.context.fill()
        #self.context.set_line_width( 3)
        #self.context.stroke()
      # text itself
      _bbox = [x1+xbearing, y1+ybearing, x1+xbearing+width, y1+ybearing+height]
      self._bboxes.append( _bbox)
      # store bbox for the first chunk only
      if not bbox or center_letter=='last':
        bbox = _bbox
      self.context.set_source_rgb( *color)
      self.context.move_to( x1, y1)
      if self.text_to_curves:
        self.context.text_path( chunk.text)
        self.context.fill()
      else:
        self.context.show_text( chunk.text)
      #self.context.fill()
      x1 += x_advance
    return bbox


  # not used
  def _draw_rectangle( self, coords, fill_color=(1,1,1)):
    #outline = self.paper.itemcget( item, 'outline')
    x1, y1, x2, y2 = coords
    self.context.set_line_join( cairo.LINE_JOIN_MITER)
    self.context.rectangle( x1, y1, x2-x1, y2-y1)
    self.context.set_source_rgb( *fill_color)
    self.context.fill_preserve()
    #self.context.set_line_width( width)
    #self.set_cairo_color( outline)
    self.context.stroke()


  def _create_cairo_path( self, points, closed=False):
    points = [self._invtransform.transform_xyz( p[0],p[1],0)[:2] for p in points]
    x, y = points[0]
    self.context.move_to( x, y)
    for (x,y) in points[1:]:
      self.context.line_to( x, y)
    if closed:
      self.context.close_path()


  def _get_bbox( self):
    bbox = list( self._bboxes[0])
    for _bbox in self._bboxes[1:]:
      x1,y1,x2,y2 = _bbox
      if x1 < bbox[0]:
        bbox[0] = x1
      if y1 < bbox[1]:
        bbox[1] = y1
      if x2 > bbox[2]:
        bbox[2] = x2
      if y2 > bbox[3]:
        bbox[3] = y2
    return bbox


  def _set_source_color( self, color):
    """depending on the value of color uses the proper method,
    either set_source_rgb or set_source_rgba"""
    if len( color) == 3:
      self.context.set_source_rgb( *color)
    elif len( color) == 4:
      self.context.set_source_rgba( *color)
    else:
      raise ValueError( "wrong specification of color '%s'" % color)



def mol_to_png( mol, filename, **kw):
  c = cairo_out( **kw)
  c.mol_to_cairo( mol, filename)


def mols_to_png( mols, filename, **kw):
  c = cairo_out( **kw)
  c.mols_to_cairo( mols, filename)


def mol_to_cairo( mol, filename, format, **kw):
  c = cairo_out( **kw)
  c.mol_to_cairo( mol, filename, format=format)


def mols_to_cairo( mols, filename, format, **kw):
  c = cairo_out( **kw)
  c.mols_to_cairo( mols, filename, format=format)



if __name__ == "__main__":

  from . import smiles

  mol = smiles.text_to_mol( "FCCSCl", calc_coords=30)
  #mol.vertices[0].properties_['show_hydrogens'] = False
  #mol.vertices[1].properties_['show_symbol'] = False
  #mol.vertices[2].properties_['show_symbol'] = True
  mol_to_png( mol, "output.png", show_hydrogens_on_hetero=True, scaling=2)

##   from . import inchi
##   mol = inchi.text_to_mol( "1/C7H6O2/c8-7(9)6-4-2-1-3-5-6/h1-5H,(H,8,9)", include_hydrogens=False, calc_coords=30)
##   mol_to_png( mol, "output.png")




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------



cdml_to_smiles = {'Me': 'C',
                  'Ph': 'c1ccccc1',
                  'Et': 'CC',
                  'OCH3': 'OC',
                  'NO2': 'N(=O)=O',
                  'COOH': 'C(=O)O',
                  'COOCH3': 'C(=O)OC',
                  'CN': 'C#N',
                  'SO3H': 'S(=O)(=O)O',
                  'PPh3': 'P(c1ccccc1)(c2ccccc2)c3ccccc3',
                  'OMe': 'OC',
                  'COCl': 'C(=O)Cl',
                  'CH2OH': 'CO'
                  }


name_to_smiles = {'Me': 'C',
                  'Ph': 'c1ccccc1',
                  'Et': 'CC',
                  'Pr': 'CCC',
                  'Bn': 'Cc1ccccc1',
                  'Bz': 'C(=O)c1ccccc1',
                  'Ac': 'C(=O)C',
                  'Ms': 'S(=O)(=O)C',
                  'Tf': 'S(=O)(=O)C(F)(F)F',
                  'Ts': 'S(=O)(=O)c1ccc(C)cc1',
                  'Bu': 'CCCC',

                  'iPr':   'C(C)C',
                  'tBu':   'C(C)(C)C',
                  'Boc':   'C(=O)OC(C)(C)C',
                  'Alloc': 'C(=O)OCC=C',
                  'Z':     'C(=O)OCc1ccccc1',
                  'Fmoc':  'C(=O)OCC1c2ccccc2-c3c1cccc3',
#                  'DMAM':  '
                  'CE':    'CCC#N',
                  'MOP':   'CC1=CC(OC)=CC=N1=O',
                  'Tr':    'C(c1ccccc1)(c2ccccc2)c3ccccc3',
                  'MMTr':  'C(c1ccc(OC)cc1)(c2ccccc2)c3ccccc3',
                  'DMTr':  'C(c1ccc(OC)cc1)(c2ccc(OC)cc2)c3ccccc3',
                  'TMS':   'Si(C)(C)C',
                  'TBDMS': 'Si(C)(C)C(C)(C)C',
                  'TBDPS': 'Si(C(C)(C)C)(c1ccccc1)c2ccccc2',
                  
                  }



#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

"""This module is intended for integration of OpenBabel/Pybel into OASA.
It provides functionality for translation of OASA molecules into Pybel molecules."""

from __future__ import print_function

import pybel, openbabel

from .atom import atom
from .bond import bond
from .molecule import molecule
from .periodic_table import periodic_table as PT



num2symbol = dict((v['ord'], k) for k, v in PT.items())


class PybelConverter( object):

  ## -------------------- Pybel to OASA --------------------

  @classmethod
  def pybel_to_oasa_atom( self, patom):
    oatom = atom()
    psymbol = num2symbol[patom.atomicnum]
    oatom.symbol = psymbol
    oatom.x, oatom.y, oatom.z = patom.coords
    oatom.charge = patom.formalcharge
    return oatom

  @classmethod
  def pybel_to_oasa_bond( self, pbond):
    obond = bond()
    obond.order = pbond.GetBondOrder()
    return obond

  @classmethod
  def pybel_to_oasa_molecule_with_atom_map( self, pmol):
    omol = molecule()
    patom_idx2oatom = {}
    for pa in pmol.atoms:
      oa = self.pybel_to_oasa_atom( pa)
      omol.add_vertex( oa)
      patom_idx2oatom[ pa.idx] = oa
    for pb in openbabel.OBMolBondIter( pmol.OBMol):
      ob = self.pybel_to_oasa_bond( pb)
      i1 = pb.GetBeginAtomIdx()
      i2 = pb.GetEndAtomIdx()
      oa1 = patom_idx2oatom[ i1]
      oa2 = patom_idx2oatom[ i2]
      omol.add_edge( oa1, oa2, ob)
    return omol, patom_idx2oatom

  @classmethod
  def pybel_to_oasa_molecule( self, *args):
    omol, patom_idx2oatom = self.pybel_to_oasa_molecule_with_atom_map( *args)
    return omol

  ## // -------------------- Pybel to OASA --------------------

  ## -------------------- OASA to Pybel --------------------

  @classmethod
  def oasa_to_ob_atom( self, oatom):
    num = PT[ oatom.symbol]['ord']
    obatom = openbabel.OBAtom()
    obatom.SetAtomicNum( num)
    obatom.SetVector( oatom.x, oatom.y, oatom.z)
    obatom.SetFormalCharge( oatom.charge)
    #patom = pybel.Atom()
    #patom.atomicnum = num
    return obatom

  @classmethod
  def oasa_to_ob_bond( self, obond):
    pbond = openbabel.OBBond()
    pbond.SetBondOrder( obond.order)
    return pbond

  @classmethod
  def oasa_to_pybel_molecule_with_atom_map( self, omol):
    obmol = openbabel.OBMol()
    oatom2obatom_idx = {}
    for i,oa in enumerate( omol.atoms):
      oba = self.oasa_to_ob_atom( oa)
      oba.SetIdx( i+1)
      obmol.AddAtom( oba)
      oatom2obatom_idx[ oa] = i+1
    for ob in omol.bonds:
      oa1, oa2 = ob.vertices
      idx1 = oatom2obatom_idx[ oa1]
      idx2 = oatom2obatom_idx[ oa2]
      obmol.AddBond( idx1, idx2, ob.order)
    pmol = pybel.Molecule( obmol)
    return pmol, oatom2obatom_idx

  @classmethod
  def oasa_to_pybel_molecule( self, *args):
    omol, patom_idx2oatom = self.oasa_to_pybel_molecule_with_atom_map( *args)
    return omol

  ## // -------------------- OASA to Pybel --------------------

  ## -------------------- conversion --------------------

  @classmethod
  def read_text( self, format, text):
    """returns a list of OASA molecules from a string"""
    obc = openbabel.OBConversion()
    if not obc.SetInFormat( format):
        raise ValueError("invalid format %s" % format)
    obmol = openbabel.OBMol()
    ok = obc.ReadString( obmol, text)
    mols = []
    while ok:
      mol = self.pybel_to_oasa_molecule( pybel.Molecule( obmol))
      if len( mol.vertices) > 0:
        # throw away empty molecules - CDX seems to cause these to appear
        mols.append( mol)
      obmol = openbabel.OBMol()
      ok = obc.Read( obmol)
    return mols

  ## // -------------------- conversion --------------------

  ## -------------------- conversion support --------------------
  ## not needed - pybel provides this functionality, I just overlooked it

  @classmethod
  def get_supported_input_formats( self):
    conv = openbabel.OBConversion()
    ret = []
    for form in conv.GetSupportedInputFormat():
      form_id, desc = [x.strip() for x in form.split("--")]
      ret.append( (form_id, desc))
    return ret

  @classmethod
  def get_supported_output_formats( self):
    conv = openbabel.OBConversion()
    ret = []
    for form in conv.GetSupportedOutputFormat():
      form_id, desc = [x.strip() for x in form.split("--")]
      ret.append( (form_id, desc))
    return ret

  ## // -------------------- conversion support --------------------

class ForceFieldOptimizer( object):
  """provides the forcefield interface for oasa molecules"""

  def __init__( self, mol, forcefield_name="Ghemical"):
    self.mol = mol
    self._amol, self._oatom2patom_idx = PybelConverter.oasa_to_pybel_molecule_with_atom_map( self.mol)
    self._patom_idx2oatom = dict((v, k) for k, v in self._oatom2patom_idx.items())
    self.forcefield_name = forcefield_name
    self.ff = openbabel.OBForceField.FindForceField( self.forcefield_name)
    self.ff.Setup( self._amol.OBMol)

  def conjugate_gradients( self, step_size=10, max_steps=100):
    """partial result will be yielded after the 'step_size' number of iteration has
    been performed. When false is yielded, we did not reach convergence criteria yet.
    'max_steps' is the number of repetitions of this, so the real number of iterations
    is 'max_steps' * 'step_size'
    """
    self.ff.ConjugateGradientsInitialize( step_size * max_steps)
    for i in range( max_steps):
      do_continue = self.ff.ConjugateGradientsTakeNSteps( step_size)
      yield not do_continue
      if not do_continue:
        break

  def update_coords( self):
    self.ff.UpdateCoordinates( self._amol.OBMol)
    for patom in self._amol:
      oatom = self._patom_idx2oatom[ patom.idx]
      oatom.coords = patom.coords



if __name__ == "__main__":
  import time

  pmol = pybel.readstring("smi", "CC(=O)O")
  omol = PybelConverter.pybel_to_oasa_molecule( pmol)
  print(omol)
  import smiles
  c = smiles.converter()
  print(c.mols_to_text([omol]))

  #print(get_supported_input_formats())
  #print(get_supported_output_formats())

  if False:
    import molfile
    with open("tbu-benzen.mol", 'r') as f:
      mol = molfile.file_to_mol(f)

    print(PybelConverter.oasa_to_pybel_molecule(mol))

    t = time.time()
    ff = ForceFieldOptimizer( mol)
    gen = ff.conjugate_gradients()
    for step in gen:
      print(step, ff.ff.Energy())
      ff.update_coords()
      print([a.coords for a in mol.atoms])
    print(time.time() - t)

  a = pybel.readstring( "smi", "CCCC\nCCC")

  for mol in PybelConverter.read_text( "smi", "CCCC\nCCC"):
    print(mol)

  print("AAAA")
  with open("/zaloha/temp/3-chlorbenzoic_acid.cdx", 'r') as f:
    t = f.read()
  pmol = pybel.readstring("cdx", t)
  print("XXX", pmol)

  for mol in PybelConverter.read_text( "cdx", t):
    print(mol)




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from __future__ import print_function

import copy
import math

from . import graph
from . import common
from . import misc
from . import transform3d
from . import periodic_table as PT
from .atom import atom
from .bond import bond
from .query_atom import query_atom



class molecule(graph.graph):

  def __init__( self, vertices =[]):
    graph.graph.__init__( self, vertices=vertices)
    # aliases
    self.atoms = self.vertices
    self.bonds = self.edges
    self.stereochemistry = []


  def __str__( self):
    return "molecule, %d atoms, %d bonds" % (len( self.vertices), len( self.edges))


  def create_vertex( self, vertex_class=None):
    if not vertex_class:
      return atom()
    else:
      return vertex_class()


  def create_edge( self):
    return bond()


  def create_graph( self):
    from . import config
    return config.Config.molecule_class()


  def add_stereochemistry( self, stereo):
    self.stereochemistry.append( stereo)


  def remove_stereochemistry( self, stereo):
    if not stereo in self.stereochemistry:
      raise ValueError("cannot remove non-existent stereochemistry information")
    self.stereochemistry.remove( stereo)


  def get_stereochemistry_by_center( self, center):
    for st in self.stereochemistry:
      if st.center == center:
        return st
    return None

  # analytics

  # override of graphs method to add stereochemistry support
  def get_disconnected_subgraphs( self):
    out = graph.graph.get_disconnected_subgraphs( self)
    for part in out:
      for st in self.stereochemistry:
        if set( [ref for ref in st.references if isinstance(ref,atom)]) <= set( part.vertices):
          part.add_stereochemistry( st)
    return out


  @property
  def weight(self):
    """Molecular weight.

    """
    w = 0
    for v in self.vertices:
      w += v.weight
      if v.free_valency > 0:
        w += v.free_valency * PT.periodic_table['H']['weight']
    return w


  @property
  def charge(self):
    """Net charge of the molecule.

    """
    return sum([a.charge for a in self.vertices])


  def get_formula_dict( self):
    """returns a formula dict as defined in the periodic_table.py::formula_dict"""
    comp = PT.formula_dict()
    for a in self.atoms:
      comp += a.get_formula_dict()
    return comp


  def add_missing_hydrogens( self):
    hs = set()
    for v in copy.copy( self.vertices):
      hs |= self.add_missing_hydrogens_to_atom( v)
    return hs


  def add_missing_hydrogens_to_atom( self, v):
    hs = set()
    for i in range( v.free_valency):
      h = self.create_vertex()
      h.symbol = 'H'
      self.add_vertex( h)
      self.add_edge( h, v)
      hs.add( h)
    return hs


  def explicit_hydrogens_to_real_atoms( self, v):
    hs = set()
    for i in range( v.explicit_hydrogens):
      h = self.create_vertex()
      h.symbol = 'H'
      self.add_vertex( h)
      self.add_edge( h, v)
      hs.add( h)
    v.explicit_hydrogens = 0
    return hs


  def add_missing_bond_orders( self, retry=False):
    """retry means to try a different approach because the last one was not successful"""
    # we must at first find orphans that will force raising of valence of their neigbors
    # NO2 is a typical example
    [v.raise_valency_to_senseful_value() for v in self.vertices if v.free_valency < 0]

    #print([sum( [xxx.free_valency for xxx in x]) for x in self._gen_free_valency_connected_components()])

##     print([map( str, n) for n in [v.neighbors for v in self.vertices if v.symbol == "N"]])
##     print([v.charge for v in self.vertices if v.symbol == "N"])

    fix = True
    while fix:
      fix = False
      for v in self.vertices:
        if sum( [n.free_valency for n in v.neighbors]) < v.free_valency:
          for n in v.neighbors:
            if n.raise_valency():
              fix = True
              break
        if fix:
          break

    #print([v.valency for v in self.vertices if v.symbol == "N"])

##     if retry:
##       for ring in self.get_smallest_independent_cycles_e():
##         if not filter( lambda x: x<=0, [v.free_valency for v in self.edge_subgraph_to_vertex_subgraph( ring)]):
##           for e in ring:
##             e.order = 4
##       self.localize_aromatic_bonds()

    # now we process it
    processed = [1]
    step = 0
    while processed:
      processed = []
      step += 1
      assert step < 100 # there is something strange if we reach this number
      for b in [bo for bo in self.edges if min( [a.free_valency for a in bo.vertices])]:
        a1, a2 = b.get_vertices()
        as1 = [a for a in a1.neighbors if a.free_valency > 0]
        as2 = [a for a in a2.neighbors if a.free_valency > 0]
        if len( as1) == 1 or len( as2) == 1:
          b.order += min( [a.free_valency for a in b.vertices])
          processed.append( b)

      # we process non-bridge bonds first, so that their fixation does not
      # force unsolvable possitioning of bonds in rings
      if not processed:
        for b in self.edges:
          i = min( [a.free_valency for a in b.vertices])
          if i and not self.is_edge_a_bridge( b):
            processed = [b]
            b.order += i
            break

      if not processed:
        for b in self.edges:
          i = min( [a.free_valency for a in b.vertices])
          if i:
            processed = [b]
            b.order += i
            break


  def _gen_free_valency_connected_components( self):
    done = set()
    vs = set( copy.copy( self.vertices))
    while vs:
      now = set()
      v = vs.pop()
      if v.free_valency:
        now.add( v)
        new = set( now)
        while new:
          added = False
          now_new = set()
          for v in new:
            for n in v.neighbors:
              if n.free_valency and not n in done:
                now.add( n)
                done.add( n)
                now_new.add( n)
          new = now_new
        vs -= now
        yield now
      else:
        done.add( v)


  def mark_aromatic_bonds( self):
    sssr = self.get_smallest_independent_cycles()
    if len( sssr) > 10:
      # turn off processing of all cycles - it would be too slow, just use SSSR
      rings = sssr
    else:
      rings = self.get_all_cycles()   #self.get_smallest_independent_cycles()
    solved = [1]
    # we need to repeat it to mark such things as 'badly' drawn naphtalene (no double bond in the centre)
    while solved:
      solved = []
      for aring in rings:
        els = [self._get_atoms_possible_aromatic_electrons( a, aring) for a in aring]
        if () in els:
          continue
        for comb in common.gen_combinations_of_series( els):
          if sum( comb) % 4 == 2:
            bring = self.vertex_subgraph_to_edge_subgraph( aring)
            for b in bring:
              b.aromatic = 1
              #b.order = 4
            solved.append( aring)
            break
      for r in solved:
        rings.remove( r)


  def _get_atoms_possible_aromatic_electrons( self, at, ring):
    out = set()
    if at.charge > 0 and at.symbol not in PT.accept_cation:
      out.add( 0)
    elif at.charge < 0:
      out.add( 2)
    if at.symbol in PT.accept_cation and not at.charge > 0:
      out.add( 2)
    if at.symbol in ('B','Al'):
      out.add( 0)
    #if at.multiplicity == 2:
    #  out.add( 1)
    for b, a in at.get_neighbor_edge_pairs():
      if b.order > 1 and (a in ring or b.aromatic):
        if b.order == 4:
          # we localize aromatic bonds
          if at.get_highest_possible_free_valency() and a.get_highest_possible_free_valency():
            # we calculate free valency with all aromatic bonds as single
            out.add( 1)
        else:
          # we detect aromatic bonds
          out.add( 1)
      elif b.order == 2 and not (a in ring):
        out.add( 0)
    return tuple( out)


  def localize_aromatic_bonds( self):
    """localizes aromatic bonds (does not relocalize already localized ones),
    for those that are not aromatic but marked so
    (it is for instance possible to misuse 'cccc' in smiles to create butadiene)
    they will be properly localized but marked as non-aromatic"""
    # at first get all clusters of aromatic bonds
    for b in copy.copy( self.edges):
      if not b.aromatic:
        self.temporarily_disconnect_edge( b)
    self.temporarily_strip_bridge_edges()
    ring_clusters = map( list, [sub for sub in self.get_connected_components() if len( sub) > 1])
    self.reconnect_temporarily_disconnected_edges()
    # now proceed in localizing double bonds in each one
    for cluster in ring_clusters:
      els = [self._get_atoms_possible_aromatic_electrons( a, cluster) for a in cluster]
      if () in els:
        continue  # misuse of aromatic bonds (e.g. by smiles) or e.g. tetrahydronaphtalene
      current_bonds = self.vertex_subgraph_to_edge_subgraph( cluster)
      combinations = list( common.gen_combinations_of_series( els))
      # try most promissing combinations first (4n+2 with maximum number of 1's and maximum sum)
      combinations.sort( reverse=True, key=lambda x: sum(x)%4==2 and 1000000+1000*len([v for v in x if v==1])+sum(x) or sum(x)+1000*len([v for v in x if v==1]))
      for comb in combinations:
        if sum( comb) % 2 == 0:
          # assign the aromatic electrons to each vertex for this run
          for i,v in enumerate( cluster):
            v.properties_['arom_els'] = comb[i]
          # at first we process the bonds that are surely single (like C-S-C in thiophene)
          to_process = []
          for b in current_bonds:
            v1,v2 = b.vertices
            if v1.properties_['arom_els'] + v2.properties_['arom_els'] != 2 or \
               (b.order in (1,4) and (not v2.get_highest_possible_free_valency() or not v1.get_highest_possible_free_valency())):
              # these bonds must be single
              # either the number of electrons between the atoms is not 2 or
              # the atoms have no possible way of accomodating different than single bond
              b.order = 1
            else:
              to_process.append( b)
          to_process_atoms = self.edge_subgraph_to_vertex_subgraph( to_process)
          processed = [v for v in cluster if not v in to_process_atoms]
          # now we use the maximum_matching graph algorithm to localize the bonds
          work_graph = self.get_induced_copy_subgraph_from_vertices_and_edges( to_process_atoms, to_process, add_back_links=True)
          #return work_graph
          mate,nrex = work_graph.get_maximum_matching()
          ok = True
          if not nrex:
            # there are no exposed vertices
            for work_v, work_n in mate.items():
              if work_n != 0:
                work_e = work_v.get_edge_leading_to( work_n)
                work_e.properties_['original'].order = 2
            for e in to_process:
              if e.order == 4:
                e.order = 1
          else:
            ok = False
          if ok:
            # additional check
            for v in cluster:
              if v.properties_['arom_els'] == 1:
                for e in v.neighbor_edges:
                  if e.order == 2:
                    # this is ok
                    break
                else:
                  # this is not good
                  ok = False
                  break
          #if not (2*len( [b for b in to_process if b.order == 2])+sum([v.properties_['arom_els'] for v in cluster if v not in to_process_atoms]) == sum( comb)):
          #  ok = False
          if not ok:
            # there are some exposed vertices left
            for b in self.vertex_subgraph_to_edge_subgraph( cluster):
              b.order = 4
          else:
            break
      if [b for b in current_bonds if b.order == 4]:
        # we did not find a matching
        raise ValueError( "Localization of aromatic bonds failed")
    self.reconnect_temporarily_disconnected_edges()
    self.localize_fake_aromatic_bonds()


  def localize_aromatic_bonds_old( self):
    """localizes aromatic bonds (does not relocalize already localized ones),
    for those that are not aromatic but marked so
    (it is for instance possible to misuse 'cccc' in smiles to create butadiene)
    they will be properly localized but marked as non-aromatic"""
    erings = self.get_smallest_independent_cycles_e()
    # filter off rings without aromatic bonds
    erings = filter( lambda x: len( [b for b in x if b.aromatic]), erings)
    rings = map( self.edge_subgraph_to_vertex_subgraph, erings)
    # sort rings
    rings.sort( lambda x,y: len(y)%2 - len(x)%2) # odd size rings first
    last_rings = []
    while rings:
      # we have to continue with the neighbor rings of the last_ring (or the rings before)
      intersection = None
      if last_rings:
        aring = None
        while last_rings:
          last_ring = last_rings.pop( -1)
          for ring in rings:
            intersection = ring & last_ring
            if intersection:
              aring = ring
              last_rings.append( last_ring)
              break
          if aring:
            break
        if not aring:
          aring = rings.pop(0)
        else:
          rings.remove( aring)
      else:
        aring = rings.pop(0)
      last_rings.append( aring)
      # convert ring from set to list
      aring = self.sort_vertices_in_path( aring, start_from=intersection and intersection.pop() or None)
      # taken from mark_aromatic_bonds
      els = [self._get_atoms_possible_aromatic_electrons( a, aring) for a in aring]
      if () in els:
        continue  # misuse of aromatic bonds (e.g. by smiles) or e.g. tetrahydronaphtalene
      for comb in common.gen_combinations_of_series( els):
        if sum( comb) % 4 == 2:
          aring.append( aring[0])
          comb.append( comb[0])
          # at first we process the bonds that are surely single (like C-S-C in thiophene)
          already_set = None
          for i in range( len( aring) -1):
            if comb[i] + comb[i+1] != 2:
              # these bonds must be single
              b = aring[i].get_edge_leading_to( aring[i+1])
              b.order = 1
              already_set = i
          if already_set is not None:
            # we reorder the comb and aring to start from the already set bonds
            j = already_set + 1
            aring = aring[j:len(aring)] + aring[1:j]
            aring.insert( 0, aring[-1])
            comb = comb[j:len(comb)] + comb[1:j]
            comb.insert( 0, comb[-1])
          i = 0
          while i+1 < len( aring):
            if comb[i] + comb[i+1] == 2:
              b = aring[i].get_edge_leading_to( aring[i+1])
              assert b is not None # should be
              # to assure alternating bonds
              bs1 = [bo for bo in aring[i].neighbor_edges
                          if bo.order == 2 and bo.aromatic and bo != b]
              bs2 = [bo for bo in aring[i+1].neighbor_edges
                          if bo.order == 2 and bo.aromatic and bo != b]
              if len( bs1) == 0 and len( bs2) == 0:
                b.order = 2
              else:
                b.order = 1
            i += 1
          break
##           # ensure chemical meaning
##           if sum( comb) == 2*sum( [b.order-1 for b in self.vertex_subgraph_to_edge_subgraph( aring[:-1])]):
##             break
##           else:
##             print("AAAA")

    self.localize_fake_aromatic_bonds()


  def localize_fake_aromatic_bonds( self):
    to_go = [b for b in self.bonds if b.order == 4]

    processed = []
    for b in to_go:
      if not min( [a.free_valency for a in b.vertices]):
        b.order = 1
        processed.append( b)
    to_go = misc.difference( to_go, processed)

    while to_go:
      # find the right bond
      b = None
      for bo in to_go:
        bs1, bs2 = bo.get_neighbor_edges2()
        if not bs1 or len( [e for e in bs1 if e.order == 1]) > 0 and len( [e for e in bs1 if e.order == 2]) == 0 \
           or not bs2 or len( [e for e in bs2 if e.order == 1]) > 0 and len( [e for e in bs2 if e.order == 2]) == 0:
          b = bo
          break
        # new start for iteration
      if not b:
        for bo in self.edges:
          if not [e for e in bo.neighbor_edges if e.order != 4]:
            b = bo
            break
      if not b:
        b = to_go.pop(0)
      # the code itself
      b.order = 2
      b.aromatic = 0
      for bo in b.neighbor_edges:
        if bo.order == 4:
          bo.order = 1
          bo.aromatic = 0
      # next turn
      to_go = [b for b in self.bonds if b.order == 4]


  def remove_unimportant_hydrogens( self):
    """removes all H atoms that don't have any special properties
    or are not part of a hydrogen only molecule."""
    only_hydrogens = True
    for v in self.vertices:
      if v.symbol != "H":
        only_hydrogens = False
        break
    if not only_hydrogens:
      for v in copy.copy( self.vertices):
        if v.symbol == 'H' and v.degree <= 1 and not v.isotope:
          self.remove_vertex( v)


  def remove_all_hydrogens( self):
    """removes all H atoms"""
    for v in copy.copy( self.vertices):
      if v.symbol == 'H' and v.degree <= 1:
        self.remove_vertex( v)


  def _get_atom_distance_matrix( self, a):
    self.mark_vertices_with_distance_from( a)
    big_out = []
    i = 0
    while 1:
      out = [a.symbol_number for a in self.vertices if a.properties_['d'] == i]
      if i > 0:
        out2 = []
        for a in self.vertices:
          if a.properties_['d'] == i-1:
            out2 += a.get_hydrogen_count() * [1]
        out += out2
      if out:
        big_out.append(tuple(sorted(out)))
      else:
        return big_out
      i += 1


  def get_symmetry_unique_atoms( self):
    out = []
    vs = copy.copy( self.vertices)
    for v in vs:
      out.append( [i for i in self._get_atom_distance_matrix( v)])
    while out:
      k = vs.pop( 0)
      v = out.pop( 0)
      o = [k]
      proc_vs = []
      proc_out = []
      while out:
        k2 = vs.pop( 0)
        v2 = out.pop( 0)
        if v2 == v:
          o.append( k2)
        else:
          proc_vs.append( k2)
          proc_out.append( v2)
      yield o
      out = proc_out
      vs = proc_vs


  def number_atoms_uniquely( self):
    out = {}
    for v in self.vertices:
      out[v] = [i for i in self._get_atom_distance_matrix( v)]
      v.properties_['distance_matrix'] = out[v]
    ms = sorted(out.values())
    #self.remove_all_hydrogens()
    i = 1
    ret = []
    for m in ms:
      ret.append(list(out.keys())[list(out.values()).index(m)])
      #if ret.symbol != 'H':
      #  print("%3d: " % i, ret, ret.free_valency)
      i += 1
    return ret


  def _read_file( self, name="/home/beda/oasa/oasa/mol.graph"):
    self.vertices = []
    self.edges = set()
    with open(name, 'r') as f:
      vs = f.readline()
      for i in vs.split(' '):
        if i != '\n':
          v = self.create_vertex()
          v.symbol = i
          self.add_vertex( v)
      for l in f.readlines():
        o, a, b = l.split(' ')
        e = self.create_edge()
        e.order = int(o)
        self.add_edge(self.vertices[int(a)], self.vertices[int(b)], e=e)


  def find_longest_mostly_carbon_chain( self):
    if len( self.vertices) < 2:
      return copy.copy( self.vertices)
    ends = [v for v in self.vertices
                if v.degree == 1]
    paths = []
    for e1 in ends:
      for e2 in ends:
        if e1 != e2:
          path = self.find_path_between( e1, e2)
          paths.append( path)  # could be optimized by not storing all the paths (once we know how to weight the chains)
    # for now we don't take care of chain composition
    l = 0
    p = None
    for path in paths:
      if len( path) > l:
        l = len( path)
        p = path
    return p


  def remove_zero_order_bonds( self):
    """zero order bonds are created for example when SMILES is read"""
    zero_bonds = [e for e in self.edges if e.order == 0]
    for b in zero_bonds:
      self.disconnect_edge( b)


  # --- the fragment matching routines ---
  def select_matching_substructures( self, other, implicit_freesites=False, auto_cleanup=True):
    """select fragments that match the complete molecule 'other' and yield them
    as lists of atoms in the order of other.vertices; however when other has
    explicit hydrogens that match implicit hydrogens on self the length of the
    returned fragment might be shorter of the matched implicit hydrogens;

    IF YOU DON'T LET THE GENERATOR RUN TO THE END OR SET THE auto_cleanup TO FALSE, YOU HAVE TO
    RUN clean_after_search MANUALLY NOT TO CLUTTER THE STRUCTURE"""

    # at first decide if we need to add implicit hydrogens to self
    add_implicit = False
    for v in other.vertices:
      if (isinstance( v, atom) and v.symbol == 'H') or \
         (isinstance( v, atom) and v.explicit_hydrogens > 0) or \
         (isinstance( v, query_atom) and ('H' in v.symbols or 'R' in v.symbols)):
        add_implicit = True
        break
    if add_implicit:
      added_hs = self.add_missing_hydrogens()
      for h in added_hs:
        h.properties_['implicit_hydrogen'] = True
    # here we add explicit hydrogens to other, we call them implicit_hydrogen to make it
    # work the same way as implicit_hydrogen
    for v in other.vertices:
      for i in range( v.explicit_hydrogens):
        h = other.create_vertex()
        other.add_vertex( h)
        h.symbol = 'H'
        e = other.create_edge()
        e.order = 1
        other.add_edge( v, h, e=e)
        h.properties_['implicit_hydrogen'] = True
      v.explicit_hydrogens = 0  # make it 0 and after search put it back
    # for implicit_freesites we add the free_sites here
    if implicit_freesites:
      for v in other.vertices:
        v.properties_['old_free_sites'] = v.free_sites
        v.free_sites = v.free_valency

    # then we create the dicts for storing threads in each of the atoms
    i = 0
    for a in other.vertices:
      a.properties_['subsearch'] = {}
    for e in other.edges | self.edges:
      e.properties_['subsearch'] = {}
    # here we select the vertex to start from
    vs = [v for v in other.vertices if isinstance( v, atom)]
    sym = common.least_common_item( [v.symbol for v in vs])
    v = [v for v in vs if v.symbol == sym][0]
    for a in self.vertices:
      a.properties_['subsearch'] = {}
      if v.matches( a):
        i += 1
        a.properties_['subsearch'][i] = v
        v.properties_['subsearch'][i] = a

    # now we can proceed with the search
    yielded = set()
    for thread in self._mark_matching_threads( v, other):
      vs = [v.properties_['subsearch'][thread] for v in other.vertices]
      # for symetrical fragments we have to get rid of copies (O1=N=O2 and O2=N=O1)
      vsset = frozenset( vs)
      if vsset not in yielded:
        if self._freesites_match( other, thread):
          yield [v for v in vs if not 'implicit_hydrogen' in v.properties_.keys()]
      yielded.add( vsset)

    if auto_cleanup:
      self.clean_after_search( other)


  def clean_after_search( self, other):
    # for implicit_freesites we restore original free_sites here
    for v in other.vertices:
      if "old_free_sites" in v.properties_:
        v.free_sites = v.properties_['old_free_sites']
        del v.properties_['old_free_sites']
    # finally we remove the added implicit hydrogens
    hs = [v for v in self.vertices if 'implicit_hydrogen' in v.properties_.keys()]
    for v in hs:
      del v.properties_['implicit_hydrogen']
      self.remove_vertex( v)
    # we also remove implicit_hydrogen (that represent atom.explicit_hydrogens) on other
    # and also set the explicit_hydrogens count back to the original atoms
    for v in other.vertices:
      if 'implicit_hydrogen' not in v.properties_:
        v.explicit_hydrogens = len( [h for h in v.neighbors if 'implicit_hydrogen' in h.properties_])
    hs = [v for v in other.vertices if 'implicit_hydrogen' in v.properties_.keys()]
    for v in hs:
      del v.properties_['implicit_hydrogen']
      other.remove_vertex( v)
    for v in self.vertices + other.vertices:
      del v.properties_['subsearch']
    for e in self.edges | other.edges:
      del e.properties_['subsearch']


  def _mark_matching_threads( self, v, other):
    """v is other vertex, other is the other molecule"""
    thread = 0
    threads = list(v.properties_['subsearch'].keys())
    while threads:
      thread = min( threads)
      threads.remove( thread)

      mirror = v.properties_['subsearch'][thread]
      for e, n in v.get_neighbor_edge_pairs():
        if thread not in n.properties_['subsearch']:
          candidates = set()
          for me, mn in mirror.get_neighbor_edge_pairs():
            if thread not in mn.properties_['subsearch'] and mn.matches( n) and me.matches( e) and thread not in e.properties_['subsearch']:
              candidates.add( (mn, me, e))

          if candidates:
            if len( candidates) > 1:
              new_threads = self._spawn_thread( other, thread, len( candidates)-1)
              ths = [thread] + new_threads
            else:
              ths = [thread]
            for c, me, e in candidates:
              th = ths.pop()
              n.properties_['subsearch'][th] = c
              c.properties_['subsearch'][th] = n
              e.properties_['subsearch'][th] = me
              me.properties_['subsearch'][th] = e
            [x for x in self._mark_matching_threads( n, other)] # just make the generator run
            if thread not in v.properties_['subsearch']:
              # the thread already died
              break
            else:
              pass
          else:
            self._delete_thread( other, thread)
            break
        # for proper handling of rings we have to check also the ones that are in this thread already
        elif thread not in e.properties_['subsearch']:
          me = self.get_edge_between( mirror, n.properties_['subsearch'][thread])
          if me and e.matches( me):
            e.properties_['subsearch'][thread] = me
            me.properties_['subsearch'][thread] = e
          else:
            self._delete_thread( other, thread)
            break
        else:
          pass

      threads = [i for i in v.properties_['subsearch'].keys() if i >= thread]
      if thread in threads:
        threads.remove( thread)
        yield thread


  def _spawn_thread( self, other, thread, number):
    my_vs = [v for v in self.vertices if thread in v.properties_['subsearch']]
    my_es = [e for e in self.edges if thread in e.properties_['subsearch']]
    other_vs = [v for v in other.vertices if thread in v.properties_['subsearch']]
    other_es = [e for e in other.edges if thread in e.properties_['subsearch']]
    max_thread = max( [max( v.properties_['subsearch'].keys()) for v in other.vertices if v.properties_['subsearch']])
    for i in range( max_thread +1, max_thread +number +1, 1):
      for v in my_vs + my_es + other_vs + other_es:
        v.properties_['subsearch'][ i] = v.properties_['subsearch'][thread]
    return list(range(max_thread + 1, max_thread + number + 1, 1))


  def _delete_thread( self, other, thread):
    for v in self.vertices + other.vertices:
      try:
        del v.properties_['subsearch'][thread]
      except KeyError:
        pass


  def _freesites_match( self, other, thread):
    for v in other.vertices:
      mirror = v.properties_['subsearch'][thread]
      unmatched_ns = [n for n in mirror.neighbors if thread not in n.properties_['subsearch']] # and not n.symbol == 'H']
      # if there is more unmatched neighbors then free-site it does not match (Hs don't count)
      if not len( unmatched_ns)+mirror.explicit_hydrogens <= v.free_sites:
        return False
    return True


  def contains_substructure( self, other, implicit_freesites=True):
    found = False
    for i in self.select_matching_substructures( other, implicit_freesites=implicit_freesites, auto_cleanup=False):
      found = True
      break
    self.clean_after_search( other)
    return found


  # // --- end of the fragment matching routines ---
  def detect_stereochemistry_from_coords( self, omit_rings=True):
    from .  import stereochemistry, geometry
    def add_neighbor_double_bonds( bond, path):
      for _e in bond.neighbor_edges:
        if _e.order == 2 and _e not in path:
          path.append( _e)
          add_neighbor_double_bonds( _e, path)
    # double bonds
    # detect clusters of double bonds
    double_paths = []
    processed = set()
    for e in self.edges:
      if e.order == 2 and e not in processed:
        if omit_rings and not self.is_edge_a_bridge( e):
          continue
        path = [e]
        add_neighbor_double_bonds( e, path)
        if len( path) % 2:
          double_paths.append( path)
          processed |= set( path)
    # detect config on these paths
    for path in double_paths:
        vertices = []
        for bond in path:
          vertices.extend( bond.vertices)
        ends = [v for v in vertices if vertices.count(v) == 1]
        if len( ends) != 2: # two ends is the only thing we are prepared to handle
          continue
        end1, end2 = ends
        # set stereochemistry for all neighbors of both ends
        for e1,n1 in end1.get_neighbor_edge_pairs():
          plane1 = geometry.plane_normal_from_3_points( (n1.x,n1.y,n1.z),(end1.x,end1.y,end1.z),(end2.x,end2.y,end2.z))
          if plane1 is None:
            continue # some coords were missing
          if not e1 in path:
            for e2,n2 in end2.get_neighbor_edge_pairs():
              if not e2 in path:
                plane2 = geometry.plane_normal_from_3_points( (end1.x,end1.y,end1.z),(end2.x,end2.y,end2.z),(n2.x,n2.y,n2.z))
                #cos_angle = geometry.same_or_oposite_side( plane1, plane2)
                cos_angle = geometry.angle_between_planes( plane1, plane2)
                if cos_angle < 0:
                  value = stereochemistry.cis_trans_stereochemistry.OPPOSITE_SIDE
                else:
                  value = stereochemistry.cis_trans_stereochemistry.SAME_SIDE
                if len( path) == 1:
                  center = path[0]
                else:
                  center = None
                refs = [n1,end1,end2,n2]
                st = stereochemistry.cis_trans_stereochemistry( center=center, value=value, references=refs)
                to_remove = None
                to_add = None
                if not self.stereochemistry:
                  self.add_stereochemistry(st)
                  continue
                for st1 in self.stereochemistry:
                  if set( st1.references) == set( st.references):
                    if st.value == st1.value:
                      break
                    else:
                      to_remove = st1
                      break
                  else:
                    self.add_stereochemistry(st)
                if to_remove is not None:
                  self.remove_stereochemistry(to_remove)


  def mark_morgan( self):
    old_morgs = 0
    for v in self.vertices:
      v.properties_['new_morgan'] = v.degree #sum( [n.degree for n in v.neighbors])
    new_morgs = len( set( [v.properties_['new_morgan'] for v in self.vertices]))

    while new_morgs > old_morgs:
      for v in self.vertices:
        v.properties_['morgan'] = v.properties_['new_morgan']

      for v in self.vertices:
        v.properties_['new_morgan'] = v.properties_['morgan'] + sum( [n.properties_['morgan'] for n in v.neighbors])
      old_morgs = new_morgs
      new_morgs = len( set( [v.properties_['new_morgan'] for v in self.vertices]))

    print("morgan", old_morgs, [v.properties_['morgan'] for v in self.vertices])


  ## some geometry related things
  def normalize_bond_length( self, bond_length=30):
    """make the average bond-length be bond_length by scaling the structure up"""
    if not self.edges or len( self.vertices) < 2:
      return False

    minx = None
    maxx = None
    miny = None
    maxy = None
    # atoms
    for v in self.vertices:
      if not maxx or v.x > maxx:
        maxx = v.x
      if not minx or v.x < minx:
        minx = v.x
      if not miny or v.y < miny:
        miny = v.y
      if not maxy or v.y > maxy:
        maxy = v.y

    scale = bond_length / self.get_mean_bond_length()
    movex = (maxx+minx)/2
    movey = (maxy+miny)/2
    trans = transform3d.transform3d()
    trans.set_move( -movex, -movey, 0)
    trans.set_scaling( scale)
    trans.set_move( movex, movey, 0)
    for v in self.atoms:
      v.x, v.y, v.z = trans.transform_xyz( v.x, v.y, v.z)
    return True


  def get_mean_bond_length( self):
    """returns the mean bond length of bonds in the molecule"""
    if len( self.edges) == 0:
      return None
    bond_length = 0.0
    for e in self.edges:
      v1, v2 = e.vertices
      bond_length += math.sqrt( (v1.x-v2.x)**2 + (v1.y-v2.y)**2)
    bl = bond_length / len( self.edges)
    return bl


  def get_structure_hash( self):
    vs = self.number_atoms_uniquely()
    ret = []
    for v in vs:
      ret.append( str(v.properties_['distance_matrix']))
    res = "*".join( ret)
    import sha
    ss = sha.new()
    ss.update( res)
    return ss.hexdigest()

  def create_CIP_digraph( self, center):
    """creates a digraph according to rules described in CIP paper."""
    assert center in self.vertices
    from .graph.digraph import digraph
    dg = digraph()
    dg.add_vertex( center.copy())
    return dg


def the_right_sorting_function( t1, t2):
  for i,l in enumerate( t1):
    k = t2[i]
    for j in range( len( l)):
      if j > len( k)-1:
        return 1
      if l[j] > k[j]:
        return 1
      elif l[j] < k[j]:
        return -1
    if len( l) < len( k):
      return -1
  return 0


def equals( mol1, mol2, level=0):
  """don't forget to put all hydrogens and bond orders to the bonds
     level 1 - number of atoms and bonds,
     level 2 - the number of atoms with same symbols is the same,
     level 3 - the whole connecivity (no stereo)
     """
  # level 1
  if not level or level >= 1:
    if not len( mol1.vertices) == len( mol2.vertices):
      return False
    if not len( mol1.edges) == len( mol2.edges):
      return False
  # level 2
  if not level or level >= 2:
    symbols1 = sorted(v.symbol for v in mol1.vertices)
    symbols2 = sorted(v.symbol for v in mol2.vertices)
    if symbols2 != symbols1:
      return False
  # level 3
  if not level or level >= 3:
    vs1 = mol1.number_atoms_uniquely()
    vs2 = mol2.number_atoms_uniquely()
    for i, v1 in enumerate( vs1):
      v2 = vs2[ i]
      if v1.properties_['distance_matrix'] != v2.properties_['distance_matrix']:
        return False
  return True



#import psyco
#psyco.profile()

##################################################
# DEMO

if __name__ == '__main__':

  def main():
    g = molecule()
    g._read_file()

    for b in g.edges:
      print(g.is_edge_a_bridge( b), end='')

##     g.add_missing_hydrogens()

##     for v in g.vertices:
##   #    print(v.symbol, v.is_chiral())
##       if v.symbol:
##         ## for a in v.get_neighbors_CIP_sorted():
##         ##   print(a, [na.symbol for na in a.neighbors])
##         if v.is_chiral():
##           print(v.symbol, [na.symbol for na in v.get_neighbors_CIP_sorted()])



  import time
  t = time.time()
  main()
  print(round( 1000*(time.time()-t), 2), 'ms')

# DEMO END
##################################################



##################################################
# TODO

# azulene is wrongly localized
# 0,2 must alternate in aromatic compound
# radical has 1 el in _get_atoms_possible_aromatic_electrons
# test the equals function

##################################################




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2009 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

"""Provide basic coordinate transformations based on matrix algebra.

"""

from __future__ import division

from math import cos, sin, pi

from . import geometry



class transform(object):
  """Provide basic higher-level interface for coordinate transforms.

  """
  def __init__( self, mat = None):
    if not mat:
      self.mat = matrix( [[1,0,0],[0,1,0],[0,0,1]])
    else:
      self.mat = matrix( mat)

  def transform_xy( self, x, y):
    x1, y1, one = self.mat.get_multiplied2( [[x],[y],[1]])
    return x1[0], y1[0]

  def transform_xyz( self, x, y, z):
    """used for compatibility with transform3d, it does not work with
    z-coord at all"""
    x,y = self.transform_xy( x, y)
    return x,y,z

  def transform_4( self, list):
    "for items that have 4 coordinates (lines, rectangles) this transforms 'em all"
    x1, y1, one = self.mat.get_multiplied2( [[list[0]],[list[1]],[1]])
    x2, y2, one = self.mat.get_multiplied2( [[list[2]],[list[3]],[1]])
    return (x1[0], y1[0], x2[0], y2[0])

  def transform_list( self, l):
    ret = []
    for line in l:
      ret.append( self.transform_xy( line[0], line[1]))
    return ret

  def transform_xy_flat_list( self, coords):
    """transforms a list that cointains alternating x,y values (not list of pairs
    as self.transform_list)"""
    ret = []
    for j in range( 0, len( coords), 2):
      x, y = self.transform_xy( coords[j], coords[j+1])
      ret.append( x)
      ret.append( y)
    return ret


  def map_transform( self, l):
    "this method can be used with map() function"
    return self.transform_xy( l[0], l[1])

  def set_move( self, dx, dy):
    "add an moving step to transformation matrix"
    self.mat = matrix( mat=self.mat.get_multiplied( [[1,0,dx],[0,1,dy],[0,0,1]]))

  def set_rotation( self, angle):
    "add an rotation step to transformation matrix"
    self.mat = matrix( mat=self.mat.get_multiplied( [[cos(angle),-sin(angle),0],[sin(angle),cos(angle),0],[0,0,1]]))

  def set_scaling_xy( self, mx, my):
    "add an scaling step to transformation matrix"
    self.mat = matrix( mat=self.mat.get_multiplied( [[mx,0,0],[0,my,0],[0,0,1]]))

  def set_scaling( self, scale):
    "add an scaling step to transformation matrix, same scaling for both dimensions"
    self.mat = matrix( mat=self.mat.get_multiplied( [[scale,0,0],[0,scale,0],[0,0,1]]))

  def get_scaling( self):
    """computes the value of scaling from self"""
    x01, y01, x02, y02 = [0, 0, 100, 100]
    x11, y11, x12, y12 = self.transform_4( (x01, y01, x02, y02))
    l1 = geometry.point_distance( x01, y01, x02, y02)
    l2 = geometry.point_distance( x11, y11, x12, y12)
    return l2/l1


  def get_scaling_xy( self):
    """computes the scaling for 'x' and 'y' separately"""
    x01, y01, x02, y02 = [0, 0, 100, 100]
    x11, y11, x12, y12 = self.transform_4( (x01, y01, x02, y02))
    return (x12-x11)/(x02-x01), (y12-y11)/(y02-y01)



class matrix(object):
  """Provide common operations for matrix of 3x3 elements.

  """
  def __init__( self, mat):
    self.mat = mat

  def get_transposed( self):
    m = self.mat
    ret = [[0,0,0],[0,0,0],[0,0,0]]
    for i in range( len( m)):
      for j in range( len( m)):
        ret[j][i] = m[i][j]
    return ret

  def get_inverse( self):
    pass

  def get_multiplied2( self, multi):
    "returns matrix multiplied by multi"
    m = []
    if type( multi) != type([]):
      for i in range( len( self.mat)):
        m.append([])
        for j in range( len( self.mat)):
          m[i].append( multi*self.mat[i][j])
    else:
      for i in range( len( self.mat)):
        m.append([])
        for j in range( len( multi[0])):
          m[i].append( self.mat[i][0]*multi[0][j] + self.mat[i][1]*multi[1][j] + self.mat[i][2]*multi[2][j])
    return  m

  def get_multiplied( self, multi):
    "returns matrix multiplied by multi"
    m = []
    if type( multi) != type([]):
      for i in range( len( self.mat)):
        m.append([])
        for j in range( len( self.mat)):
          m[i].append( multi*self.mat[i][j])
    else:
      for i in range( len( multi)):
        m.append([])
        for j in range( len( self.mat)):
          m[i].append( multi[i][0]*self.mat[0][j] + multi[i][1]*self.mat[1][j] + multi[i][2]*self.mat[2][j])
    return  m

  def get_determinant( self):
    m = self.mat
    return (((m[0][0] * m[1][1] * m[2][2]) + (m[0][1] * m[1][2] * m[2][0]) + (m[0][2] * m[1][0] * m[2][1])) -
            ((m[2][1] * m[1][2] * m[0][0]) + (m[2][2] * m[1][0] * m[0][1]) + (m[2][0] * m[1][1] * m[0][2])))




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

"""Vertex class.

"""

import copy


class vertex(object):
  """Simple vertex class.

  Normaly would not be needed but it can speed up many analytical tasks
  to store data directly in vertex and not get them from the graph
  connectivity matrix.
  Vertex has a value attribute used to store arbitrary objects.
  """
  attrs_to_copy = ("value",)

  def __init__(self):
    self.properties_ = {} # used to store intermediate properties such as distances etc.
    self.value = None  # used to store any object associated with the vertex
    self._neighbors = {} # set of all neighbors
    self._clean_cache()


  def __str__(self):
    return ("vertex, value=%s, degree=%d, " % (str(self.value), self.degree)) + str(self.properties_)


  def _clean_cache(self):
    self._cache = {}


  def copy(self):
    other = self.__class__()
    for attr in self.attrs_to_copy:
      setattr(other, attr, copy.copy(getattr(self, attr)))
    return other


  def add_neighbor(self, v, e):
    """adds a neighbor connected via e"""
    self._clean_cache()
    self._neighbors[e] = v


  def remove_neighbor(self, v):
    self._clean_cache()
    to_del = None
    for k, vv in self._neighbors.items():
      if v == vv:
        to_del = k
        break
    if to_del:
      del self._neighbors[to_del]
    else:
      raise Exception("Cannot remove non-existing neighbor")


  def remove_edge_and_neighbor(self, e):
    self._clean_cache()
    if e in self._neighbors.keys():
      del self._neighbors[e]
    else:
      raise Exception("Cannot remove non-existing edge", e)


  @property
  def neighbors(self):
    """Neighboring vertices.

    """
    return [v for (e,v) in self._neighbors.items()
                  if not e.disconnected]


  def get_neighbor_connected_via( self, e):
    return self._neighbors[e]


  def get_edge_leading_to(self, a):
    for b, at in self._neighbors.items():
      if a == at:
        return b
    return None


  @property
  def degree(self):
    """Degree of the vertex.

    """
    return len(self.neighbors)


  def get_neighbors_with_distance( self, d):
    ret = []
    for v in self.neighbors:
      if 'd' in v.properties_ and v.properties_['d'] == d:
        ret.append(v)
    return ret


  def get_neighbor_edge_pairs(self):
    for e, v in self._neighbors.items():
      if not e.disconnected:
        yield e,v


  @property
  def neighbor_edges(self):
    """Neighboring edges.

    """
    return [e for e in self._neighbors.keys()
                if not e.disconnected]




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from __future__ import print_function

import re

from . import reaction
from . import oasa_exceptions
from . import stereochemistry
from . import periodic_table as PT
from .config import Config
from .plugin import plugin
from .molecule import molecule, equals



class smiles( plugin):

  name = "smiles"
  read = 1
  write = 1

  smiles_to_oasa_bond_recode = {'-': 1, '=': 2, '#': 3, ':': 4, ".": 0, "\\": 1, "/": 1}
  oasa_to_smiles_bond_recode = {1: '', 2: '=', 3: '#', 4:''}

  organic_subset = ['B', 'C', 'N', 'O', 'P', 'S', 'F', 'Cl', 'Br', 'I']

  def __init__( self, structure=None):
    self.structure = structure

  def recode_oasa_to_smiles_bond( self, b):
    if b.aromatic:
      return ''
    elif b in self._stereo_bonds_to_others:
      others = [(e,st) for e,st in self._stereo_bonds_to_others[b] if e in self._stereo_bonds_to_code]
      if not others:
        code = "\\"
      else:
        # other bonds enforce encoding of this one, we select the first one,
        # bacause there is nothing we can do if there are clashing constrains anyway
        other, st = others[0]
        end1,inside1,inside2,end2 = st.references
        if set( other.vertices) == set( [end1,inside1]):
          v1 = inside1
          v2 = end1
        else:
          v1 = inside2
          v2 = end2
        last_order = self._processed_atoms.index( v2) - self._processed_atoms.index( v1)
        last_code = self._stereo_bonds_to_code[ other] == "\\" and 1 or -1
        relation = st.value == st.OPPOSITE_SIDE and -1 or 1
        if relation*last_code*last_order < 0:
          code = "/"
        else:
          code = "\\"
      self._stereo_bonds_to_code[ b] = code
      return code
    else:
      if b.order == 1:
        a1, a2 = b.vertices
        if 'aromatic' in a1.properties_ and 'aromatic' in a2.properties_:
          # non-aromatic bond connecting two aromatic rings, we need to return -
          return '-'
      return self.oasa_to_smiles_bond_recode[ b.order]

  def set_structure( self, structure):
    self.structure = structure

  def get_structure( self):
    return self.structure

  def read_smiles( self, text, explicit_hydrogens_to_real_atoms=False):
    self.explicit_hydrogens_to_real_atoms = explicit_hydrogens_to_real_atoms
    mol = Config.create_molecule()
    text = "".join( text.split())
    is_text = re.compile("^[A-Z][a-z]?$")
    # internally revert \/ bonds before numbers, this makes further processing much easier
    text = re.sub( r"([\\/])([0-9])", lambda m: (m.group(1)=="/" and "\\" or "/")+m.group(2), text)
    # // end
    chunks = re.split( "(\[.*?\]|[A-Z][a-z]?|%[0-9]{1,2}|[^A-Z]|[a-z])", text)
    chunks = self._check_the_chunks( chunks)
    last_atom = None
    last_bond = None
    numbers = {}
    bracket_openings = []
    for c in chunks:
      # atom
      if is_text.match( c) or c.islower() or c[0] == "[":
        a = mol.create_vertex()
        if c[0] == "[":
          # atom spec in square brackets
          self._parse_atom_spec( c, a)
        else:
          # just atom symbol
          if c.islower():
            symbol = c.capitalize()
            a.properties_['aromatic'] = 1
          else:
            symbol = c
          a.symbol = symbol

        mol.add_vertex( a)
        if last_bond: # and not (not 'aromatic' in a.properties_ and last_bond.aromatic):
          # make last bond aromatic if it was stereo and atoms are aromatic
          if 'stereo' in last_bond.properties_ and \
             'aromatic' in last_atom.properties_ and \
             'aromatic' in a.properties_ and \
             not last_bond.aromatic:
            last_bond.aromatic = True
          mol.add_edge( last_atom, a, e=last_bond)
          last_bond = None
        elif last_atom:
          b = mol.add_edge( last_atom, a)
          if 'aromatic' in a.properties_:
            # aromatic bond
            b.order = 4
            b.type = 'n'
        last_atom = a
        last_bond = None
      # bond
      elif c in r'-=#:.\/':
        order = self.smiles_to_oasa_bond_recode[ c]
        last_bond = mol.create_edge()
        last_bond.order = order
        last_bond.type = 'n'
        if c in r'\/':
          last_bond.properties_['stereo'] = c
      # ring closure
      elif c.isdigit():
        if c in numbers:
          if last_bond:
            b = last_bond
          else:
            b = mol.create_edge()
            if "aromatic" in numbers[c].properties_:
              b.order = 4
          mol.add_edge( last_atom, numbers[c], e=b)
          last_bond = None
          del numbers[ c]
        else:
          numbers[c] = last_atom
          last_bond = None
      elif c == '(':
        bracket_openings.append( last_atom)
      elif c == ')':
        last_atom = bracket_openings.pop(-1)

    ## FINISH
    # deal with explicit valency, etc.
    for a in mol.vertices:
      if not "explicit_valency" in a.properties_:
        a.raise_valency_to_senseful_value()
      else:
        # detect radicals (but not biradicals - problem of triplet vs. singlet)
        del a.properties_['explicit_valency']
        if a.valency - a.occupied_valency == 1:
          a.multiplicity += 1
        else:
          a.valency = a.occupied_valency
      try:
        del a.properties_['aromatic']
      except:
        pass

    # stereochemistry
    self._process_stereochemistry( mol)

    if len(mol.vertices) == 0:
      mol = None
    self.structure = mol


  def _parse_atom_spec( self, c, a):
    """c is the text spec,
    a is an empty prepared vertex (atom) instance"""
    bracketed_atom = re.compile("^\[(\d*)([A-z][a-z]?)(.*?)\]")
    m = bracketed_atom.match( c)
    if m:
      isotope, symbol, rest = m.groups()
    else:
      raise ValueError( "unparsable square bracket content '%s'" % c)
    if symbol.islower():
      symbol = symbol.capitalize()
      a.properties_['aromatic'] = 1
    a.symbol = symbol
    if isotope:
      a.isotope = int( isotope)
    # hydrogens
    _hydrogens = re.search( "H(\d*)", rest)
    h_count = 0
    if _hydrogens:
      if _hydrogens.group(1):
        h_count = int( _hydrogens.group(1))
      else:
        h_count = 1
    a.explicit_hydrogens = h_count
    # charge
    charge = 0
    # one possible spec of charge
    _charge = re.search( "[-+]{2,10}", rest)
    if _charge:
      charge = len( _charge.group(0))
      if _charge.group(0)[0] == "-":
        charge *= -1
    # second one, only if the first one failed
    else:
      _charge = re.search( "([-+])(\d?)", rest)
      if _charge:
        if _charge.group(2):
          charge = int( _charge.group(2))
        else:
          charge = 1
        if _charge.group(1) == "-":
          charge *= -1
    a.charge = charge
    # stereo
    _stereo = re.search( "@+", rest)
    if _stereo:
      stereo = _stereo.group(0)
      a.properties_['stereo'] = stereo
    # using [] means valency is explicit
    a.properties_['explicit_valency'] = True

  def _check_the_chunks( self, chunks):
    is_text = re.compile("^[A-Z][a-z]?$")
    is_long_num = re.compile( "^%[0-9]{1,2}$")
    ret = []
    for c in chunks:
      if not c:
        continue
      if is_text.match( c):
        if (not c in PT.periodic_table and len(c)==2) or c == "Sc": # Sc is S-c not scandium
          a,b = c
          ret.append( a)
          ret.append( b)
        else:
          ret.append( c)
      elif is_long_num.match( c):
        ret.append( str( int( c[1:])))
      else:
        ret.append( c)
    return ret



  def _process_stereochemistry( self, mol):
    ## process stereochemistry
    ## double bonds
    def get_stereobond_direction( end_atom, inside_atom, bond, init):
      position = mol.vertices.index( end_atom) - mol.vertices.index( inside_atom)
      char = bond.properties_['stereo'] == "\\" and 1 or -1
      direction = (position * char * init) < 0 and "up" or "down"
      return direction
    def get_end_and_inside_vertex_from_edge_path( edge, path):
      a1,a2 = edge.vertices
      if len( [e for e in a1.neighbor_edges if e in path]) == 1:
        return a1, a2
      return a2, a1

    stereo_edges = [e for e in mol.edges if "stereo" in e.properties_]
    paths = []
    for i,e1 in enumerate( stereo_edges):
      for e2 in stereo_edges[i+1:]:
        path = mol.get_path_between_edges( e1, e2)
        path2 = path[1:-1]
        if len( path2)%2 and not [_e for _e in path2 if _e.order != 2]:
          # only odd number of double bonds, double bonds only
          for _e in path[1:-1]:
            if not mol.is_edge_a_bridge_fast_and_dangerous( _e):
              break
          else:
            # only stereo related to non-cyclic bonds
            paths.append( path)

    for path in paths:
      bond1 = path[0]
      end_atom1,inside_atom1 = get_end_and_inside_vertex_from_edge_path( bond1, path)
      bond2 = path[-1]
      end_atom2,inside_atom2 = get_end_and_inside_vertex_from_edge_path( bond2, path)
      d1 = get_stereobond_direction( end_atom1, inside_atom1, bond1, -1)
      d2 = get_stereobond_direction( end_atom2, inside_atom2, bond2, -1)
      if d1 == d2:
        value = stereochemistry.cis_trans_stereochemistry.SAME_SIDE
      else:
        value = stereochemistry.cis_trans_stereochemistry.OPPOSITE_SIDE
      if len( path) == 3:
        center = path[1]
      else:
        center = None
      refs = [end_atom1,inside_atom1,inside_atom2,end_atom2]
      st = stereochemistry.cis_trans_stereochemistry( center=center, value=value, references=refs)
      mol.add_stereochemistry( st)

    # tetrahedral stereochemistry
    for v in mol.vertices:
      refs = None
      if 'stereo' in v.properties_:
        idx = sorted(mol.vertices.index(n) for n in v.neighbors)
        if len( idx) < 3:
          pass # no stereochemistry with less then 3 neighbors
        elif len( idx) == 3:
          if v.explicit_hydrogens == 0:
            pass # no stereochemistry without adding hydrogen here
          else:
            if self.explicit_hydrogens_to_real_atoms:
              hs = mol.explicit_hydrogens_to_real_atoms( v)
              h = hs.pop()
            else:
              h = stereochemistry.explicit_hydrogen()
            v_idx = mol.vertices.index( v)
            idx1 = [i for i in idx if i < v_idx]
            idx2 = [i for i in idx if i > v_idx]
            refs = [mol.vertices[i] for i in idx1] + [h] + [mol.vertices[i] for i in idx2]
        elif len( idx) == 4:
          refs = [mol.vertices[i] for i in idx]
        else:
          pass # unhandled stereochemistry
      if refs:
        if v.properties_["stereo"] == "@":
          direction = stereochemistry.tetrahedral_stereochemistry.ANTICLOCKWISE
        elif v.properties_['stereo'] == "@@":
          direction = stereochemistry.tetrahedral_stereochemistry.CLOCKWISE
        else:
          continue # no meaning
        st = stereochemistry.tetrahedral_stereochemistry( center=v, value=direction, references=refs)
        mol.add_stereochemistry( st)

    # delete the data after processing
    for e in mol.edges:
      if 'stereo' in e.properties_:
        del e.properties_['stereo']
    for v in mol.vertices:
      if 'stereo' in v.properties_:
        del v.properties_['stereo']


  def get_smiles( self, mol):
    if not mol.is_connected():
      raise oasa_exceptions.oasa_not_implemented_error( "SMILES", "Cannot encode disconnected compounds, such as salts etc. HINT - use molecule.get_disconnected_subgraphs() to divide the molecule to individual parts.")
    #mol = molec.copy()
    self.molecule = mol
    self.ring_joins = []
    self._processed_atoms = []
    self.branches = {}
    self._stereo_bonds_to_code = {} # for bond it will contain character it uses
    self._stereo_bonds_to_others = {} # for bond it will contain the other bonds
    self._stereo_centers = {}
    # at first we mark all the atoms with aromatic bonds
    # it is much simple to do it now when all the edges are present
    # we can make use of the properties attribute of the vertex
    for e in mol.edges:
      if e.aromatic:
        for v in e.vertices:
          v.properties_[ 'aromatic'] = 1
    # stereochemistry information preparation
    for st in mol.stereochemistry:
      if isinstance( st, stereochemistry.cis_trans_stereochemistry):
        end1, inside1, inside2, end2 = st.references
        e1 = end1.get_edge_leading_to( inside1)
        e2 = end2.get_edge_leading_to( inside2)
        self._stereo_bonds_to_others[ e1] = self._stereo_bonds_to_others.get( e1, []) + [(e2, st)]
        self._stereo_bonds_to_others[ e2] = self._stereo_bonds_to_others.get( e2, []) + [(e1, st)]
      elif isinstance( st, stereochemistry.tetrahedral_stereochemistry):
        self._stereo_centers[st.center] = st
      else:
        pass # we cannot handle this

    ret = ''.join( [i for i in self._get_smiles( mol)])
    mol.reconnect_temporarily_disconnected_edges()
    # this is needed because the way temporarily_disconnected edges are handled is not compatible with the way smiles
    # generation works - it splits the molecule while reusing the same atoms and bonds and thus disconnected bonds accounting fails
    for e in mol.edges:
      e.disconnected = False
    # here tetrahedral stereochemistry is added
    for v, st in self._stereo_centers.items():
      processed_neighbors = []
      for n in self._processed_atoms:
        if n in v.neighbors:
          processed_neighbors.append( n)
        elif v.explicit_hydrogens and n is v:
          processed_neighbors.append( stereochemistry.explicit_hydrogen())
      count = match_atom_lists( st.references, processed_neighbors)
      clockwise = st.value == st.CLOCKWISE
      if count % 2 == 1:
        clockwise = not clockwise
      ch_symbol = clockwise and "@@" or "@"
      ret = ret.replace( "{{stereo%d}}" % mol.vertices.index(v), ch_symbol)
    return ret



  def _get_smiles( self, mol, start_from=None):
    # single atoms
    if len( mol.vertices) == 1:
      v = mol.vertices[0]
      yield self._create_atom_smiles( v)
      for e in self.ring_joins:
        if v in e.get_vertices():
          yield self.recode_oasa_to_smiles_bond( e)
          yield self._create_ring_join_smiles( self.ring_joins.index( e))
      return
    while not (is_line(mol) and (not start_from or start_from.degree <= 1)):
      if is_pure_ring( mol):
        if start_from:
          self.ring_joins.append( mol.temporarily_disconnect_edge( start_from.neighbor_edges[0]))
        else:
          self.ring_joins.append( mol.temporarily_disconnect_edge( list( mol.edges)[0]))
      else:
        e, mol, branch_vertex, branch = self.disconnect_something( mol, start_from=start_from)
        if branch_vertex:
          if branch_vertex in self.branches:
            self.branches[ branch_vertex].append((e, branch))
          else:
            self.branches[ branch_vertex] = [(e, branch)]
        else:
          self.ring_joins.append( e)
    try:
      start, end = filter(lambda x: x.degree == 1, mol.vertices)
    except:
      #print(filter(lambda x: x.degree == 1, mol.vertices))
      raise Exception("shit")
    if start_from == end:
      start, end = end, start
    v = start
    last = None
    was_end = False
    while True:
      yield self._create_atom_smiles( v)
      # the atom
      for e in self.ring_joins:
        if v in e.get_vertices():
          _b = self.recode_oasa_to_smiles_bond( e)
          if _b not in "/\\":
            yield _b
          yield self._create_ring_join_smiles( self.ring_joins.index( e))
      # branches
      if v in self.branches:
        for edg, branch in self.branches[ v]:
          yield '('
          yield self.recode_oasa_to_smiles_bond( edg)
          v1, v2 = edg.vertices
          vv = (v1 != v) and v1 or v2
          for i in self._get_smiles( branch, start_from=vv):
            yield i
          yield ')'
      # bond leading to the neighbor
      for e, neighbor in v.get_neighbor_edge_pairs():
        if neighbor != last:
          yield self.recode_oasa_to_smiles_bond( e)
          last = v
          v = neighbor
          break
      if was_end:
        break
      if v == end:
        was_end = True


  def _create_atom_smiles( self, v):
    self._processed_atoms.append( v)
    if 'aromatic' in v.properties_.keys():
      symbol = v.symbol.lower()
    else:
      symbol = v.symbol

    stereo = self._stereo_centers.get( v, None)

    if (v.symbol not in self.organic_subset) or (v.isotope) or (v.charge != 0) or (v.valency != PT.periodic_table[ v.symbol]['valency'][0]) or (stereo) or (v.multiplicity != 1):
      # we must use square bracket
      isotope = v.isotope and str( v.isotope) or ""
      # charge
      if v.charge:
        sym = v.charge < 0 and "-" or "+"
        charge = sym + (abs( v.charge) > 1 and str( abs( v.charge)) or "")
      else:
        charge = ""
      # explicit hydrogens
      num_h = v.valency - v.occupied_valency + v.explicit_hydrogens
      h_spec = (num_h and "H" or "") + (num_h > 1 and str( num_h) or "")
      # stereo
      if stereo:
        stereo = "{{stereo%d}}" % self.molecule.vertices.index( v)
      else:
        stereo = ""
      return "[%s%s%s%s%s]" % (isotope, symbol, stereo, h_spec, charge)
    else:
      # no need to use square brackets
      return symbol


  def disconnect_something( self, mol, start_from=None):
    """returns (broken edge, resulting mol, atom where mol was disconnected, disconnected branch)"""
    # we cannot do much about this part
    if start_from and start_from.degree != 1:
      for e,n in start_from.get_neighbor_edge_pairs():
        if n.degree > 2:
          mol.temporarily_disconnect_edge( e)
          return e, mol, None, None
      mol.temporarily_disconnect_edge( e)
      return e, mol, None, None
    # at first try to find edges for which degree of neighbors is bigger
    # than [2,2] and at best they are not bridges
    # when no non-bridges are present use the other ones
    #
    # the edges with crowded atoms
    for e in mol.edges:
      d1, d2 = [x.degree for x in e.get_vertices()]
      if d1 > 2 and d2 > 2 and not mol.is_edge_a_bridge_fast_and_dangerous( e):
        mol.temporarily_disconnect_edge( e)
        return e, mol, None, None
    # the other valuable non-bridge edges
    for e in mol.edges:
      d1, d2 = [x.degree for x in e.get_vertices()]
      if (d1 > 2 or d2 > 2) and not mol.is_edge_a_bridge_fast_and_dangerous( e):
        mol.temporarily_disconnect_edge( e)
        return e, mol, None, None
    # there are no non-bridges
    # we want to split off the smallest possible chunks
    min_size = None
    the_right_edge = None
    the_right_mol = None
    the_right_branch = None
    the_right_branch_atom = None
    ring_joints_in_branch = 1000
    ring_join_vertices = set(j for i in (e.vertices for e in self.ring_joins)
                                   for j in i)
    for e in mol.edges:
      d1, d2 = [x.degree for x in e.get_vertices()]
      if d1 > 2 or d2 > 2:
        ps = mol.get_pieces_after_edge_removal( e)
        if len( ps) == 1:
          print("impossible")
          continue
        lenghts = map( len, ps)
        ms = min( lenghts)
        p1, p2 = ps
        the_mol = (len( p1) < len( p2)) and p2 or p1
        the_branch = (p1 == the_mol) and p2 or p1
        ring_joints = len( [i for i in the_branch if i in ring_join_vertices])
        if not min_size or ms < min_size or ring_joints_in_branch > ring_joints:
          min_size = ms
          the_right_edge = e
          the_right_mol = the_mol
          the_right_branch = the_branch
          ring_joints_in_branch = ring_joints
    if the_right_edge:
      # what is possible to make here instead in the loop is made here
      # it saves time
      v1, v2 = the_right_edge.vertices
      mol.temporarily_disconnect_edge( the_right_edge)
      the_right_branch_atom = (v1 in the_right_mol) and v1 or v2
      the_right_mol = mol.get_induced_subgraph_from_vertices( the_right_mol)
      the_right_branch = mol.get_induced_subgraph_from_vertices( the_right_branch)
      return (the_right_edge,
              the_right_mol,
              the_right_branch_atom,
              the_right_branch)
    #print(mol, mol.is_connected())
    raise Exception("fuck, how comes!?")




  def disconnect_something_simple( self, mol, start_from=None):
    """returns (broken edge, resulting mol, atom where mol was disconnected, disconnected branch)"""
    # we cannot do much about this part
    if not mol.is_connected():
      print("unconnected ", mol)
    if start_from and start_from.degree > 1:
      e = list(start_from._neighbors.keys())[0]
      mol.disconnect_edge( e)
      ps = [i for i in mol.get_connected_components()]
      if len( ps) == 1:
        return e, mol, None, None
      else:
        p1 = mol.get_induced_subgraph_from_vertices( ps[0])
        p2 = mol.get_induced_subgraph_from_vertices( ps[1])
        v1, v2 = e.get_vertices()
        p = (start_from in p1.vertices) and p1 or p2
        px = (p1 == p) and p2 or p1
        v = (v1 in p.vertices) and v1 or v2
        return e, p, v, px

    for e in mol.edges:
      d1, d2 = [x.degree for x in e.get_vertices()]
      if (d1 > 2 or d2 > 2):
        mol.disconnect_edge( e)
        ps = [i for i in mol.get_connected_components()]
        if len( ps) == 1:
          return e, mol, None, None
        else:
          p1 = mol.get_induced_subgraph_from_vertices( ps[0])
          p2 = mol.get_induced_subgraph_from_vertices( ps[1])
          v1, v2 = e.get_vertices()
          if start_from:
            p = (start_from in p1.vertices) and p1 or p2
            px = (p1 == p) and p2 or p1
            v = (v1 in p.vertices) and v1 or v2
            return e, p, v, px

          v = (v1 in p1.vertices) and v1 or v2
          return e, p1, v, p2
    print(mol, mol.is_connected(), ',', map( len, mol.get_connected_components()), ',', start_from)
    raise Exception("fuck, how comes!?")

  @staticmethod
  def _create_ring_join_smiles( index):
    i = index +1
    if i > 9:
      return "%%%d" % i
    else:
      return str( i)



def is_line( mol):
  """all degrees are 2 except of two with degree 1"""
  if len( mol.vertices) == 1:
    return True
  ones = 0
  for v in mol.vertices:
    d = v.degree
    if d == 1:
      if ones == 2:
        return False
      ones += 1
    elif d != 2:
      return False
  if ones == 2:
    return True
  return False


def is_pure_ring(mol):
  return list(filter(lambda x: x.degree != 2, mol.vertices)) == []


def match_atom_lists( l1, l2):
  """sort of bubble sort with counter"""
  count = 0
  for i1, v1 in enumerate( l1):
    for i2 in range( i1, len( l2)):
      v2 = l2[i2]
      if v2 == v1:
        if i1 != i2:
          l2[i1],l2[i2] = l2[i2],l2[i1]
          count += 1
        break
  assert l1 == l2
  return count


##################################################
## MODULE INTERFACE - newstyle

from .converter_base import converter_base

class smiles_converter( converter_base):

  # standard converter attrs
  reads_text = True
  writes_text = True
  reads_files = True
  writes_files = True

  default_configuration = {"R_GENERATE_COORDS": True,
                           "R_BOND_LENGTH": 1,
                           "R_LOCALIZE_AROMATIC_BONDS": True,
                           "R_EXPLICIT_HYDROGENS_TO_REAL_ATOMS": False,

                           "W_AROMATIC_BOND_AUTODETECT": True,
                           "W_INDIVIDUAL_MOLECULE_SEPARATOR": ".",
                           "W_DETECT_STEREO_FROM_COORDS": True,
                           }

  def __init__( self):
    converter_base.__init__( self)

  def mols_to_text( self, structures):
    converter_base.mols_to_text( self, structures)
    sm = smiles()
    ret = []
    for mol in structures:
      if self.configuration["W_AROMATIC_BOND_AUTODETECT"]:
        mol.mark_aromatic_bonds()
      if self.configuration["W_DETECT_STEREO_FROM_COORDS"]:
        mol.detect_stereochemistry_from_coords()
      ret.append( sm.get_smiles( mol))
    self.last_status = self.STATUS_OK
    return self.configuration["W_INDIVIDUAL_MOLECULE_SEPARATOR"].join( ret)

  def read_text( self, text):
    # reaction support
    if ">" in text:
      if text.count(">") != 2:
        raise oasa_exceptions.oasa_smiles_error( "Only two '>' characters are supported in SMILES reactions. %d present" % text.count(">"))
      parts = text.split( ">")
      res = []
      for part in parts:
        res.append( self._read_string( part))
        if not self.last_status == self.STATUS_OK:
          res = []
          break
      if not res:
        self.result = []
        return []
      else:
        react = reaction.reaction()
        for p in res[0]:
          react.reactants.append( reaction.reaction_component( p))
        for p in res[2]:
          react.products.append( reaction.reaction_component( p))
        for p in res[1]:
          react.reagents.append( reaction.reaction_component( p))
        self.result = [react]
        return self.result
    else:
      return self._read_string( text)

  def _read_string( self, text):
    converter_base.read_text( self, text)
    sm = smiles()
    sm.read_smiles( text, explicit_hydrogens_to_real_atoms=self.configuration['R_EXPLICIT_HYDROGENS_TO_REAL_ATOMS'])
    mol = sm.structure
    if mol is None:
      return []
    mol.remove_zero_order_bonds()
    # Keep components sorted in input order
    mols = sorted(mol.get_disconnected_subgraphs(),
                  key=lambda c: mol.vertices.index(c.vertices[0]))
    for mol in mols:
      if self.configuration["R_LOCALIZE_AROMATIC_BONDS"]:
        mol.localize_aromatic_bonds()
        for b in mol.bonds:
          b.aromatic = 0
      if self.configuration["R_GENERATE_COORDS"]:
        coords_generator.calculate_coords( mol, bond_length=self.configuration['R_BOND_LENGTH'])
    self.result = mols
    self.last_status = self.STATUS_OK
    return mols

  def mols_to_file( self, structures, f):
    converter_base.mols_to_file( self, structures, f)
    f.write( self.mols_to_text( structures))

  def read_file( self, f):
    converter_base.read_file( self, f)
    mols = []
    for line in f:
      mol = self.text_to_mols( line)
      mols.extend( mol)
    return mols

converter = smiles_converter

# END OF MODULE INTERFACE
##################################################

##################################################
## MODULE INTERFACE - oldstyle

from . import coords_generator

reads_text = True
writes_text = True
reads_files = True
writes_files = True

def mol_to_text( structure):
  sm = smiles()
  structure.mark_aromatic_bonds()
  return sm.get_smiles( structure)

def text_to_mol( text, calc_coords=1, localize_aromatic_bonds=True):
  sm = smiles()
  sm.read_smiles( text)
  mol = sm.structure
  if localize_aromatic_bonds:
    mol.localize_aromatic_bonds()
    for b in mol.bonds:
      b.aromatic = 0
  if calc_coords:
    coords_generator.calculate_coords( mol, bond_length=calc_coords)
  return mol

def mol_to_file( mol, f):
  f.write( mol_to_text( mol))

def file_to_mol( f):
  return text_to_mol( f.read())


# END OF MODULE INTERFACE
##################################################



##################################################
# DEMO

if __name__ == '__main__':
  import sys
  import time

  def main( text, cycles):
    t = time.time()
    conv = converter()
    conv.configuration['R_EXPLICIT_HYDROGENS_TO_REAL_ATOMS'] = True
    for j in range( cycles):
      mols = conv.read_text( text)
      for mol in mols:
        #mol.remove_unimportant_hydrogens()
        print("  summary formula:   ", mol.get_formula_dict())
      text = conv.mols_to_text( mols)
      print("  generated SMILES:   %s" % text)
      print("  --")
    t = time.time()-t
    print('time per cycle', round( 1000*t/cycles, 2), 'ms')

  repeat = 3

  if not len( sys.argv) > 1:
    text = "COc5ccc4c2sc(cc2nc4c5)-c(cc1nc3c6)sc1c3ccc6OC"  #"ccc4ccc2cc1cc3ccccc3cc1cc2c4"
    text = "[C@@H](C)(CC)O"
    text = "O=C1c2c3c4c(cc2)c2ccc5c6c2c(ccc6C(=O)c2c5cccc2)c4ccc3c2ccccc12"
  else:
    text = sys.argv[1]

  print("oasa::SMILES DEMO")
  print("converting following smiles to smiles (%d times)" % repeat)
  print("  starting with:      %s" % text)
  print("  --------------------")
  main( text, repeat)

# DEMO END
##################################################






##################################################
# TODO

# last branch does not need to be branch
# optimize for either speed or human-readability
# at first get rid of the edges common to two or more rings, not critical
# handling of the start_from in disconnect_something as in disconnect_something_simple
# could in disconnect_something happen that a start_from will be returned in a branch?
# if I do "print(m.get_smiles( m.structure))" the ring counting does not work after that
## the transformation can be destructive!!!  - check

## THIS IS A PROBLEM : C=1ccC=2C=1C=CC=CC=2  (should be azulene)

# E/Z stereo is ignored in rings



#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from __future__ import print_function

import os, sys, re
try:
    from pysqlite2 import dbapi2 as sqlite
except ImportError as e:
    raise Exception( "The required pysqlite module could not be loaded. More info here: '%s'" % e)

from . import inchi
from . import oasa_exceptions



class Config(object):
    database_file = os.path.abspath(os.path.join(os.path.dirname(__file__), "structures.db"))



def create_database():
  connection = sqlite.connect( Config.database_file)
  c = connection.cursor()
  c.execute( "DROP TABLE IF EXISTS structures;")
  c.execute( """CREATE TABLE structures (
  id INTEGER PRIMARY KEY,
  name TEXT,
  inchikey TEXT,
  smiles TEXT);""")
  connection.commit()
  c.execute( """CREATE TABLE synonyms (
  id INTEGER,
  synonym TEXT);""")
  connection.commit()
  c.close()
  connection.close()


def normalize_inchi( inchi):
    if inchi.startswith("InChI="):
        return inchi[6:]
    else:
        return inchi

def _open_infile( infilename):
    if os.path.isfile( infilename):
        if infilename.endswith(".gz"):
            import gzip
            f = gzip.open( infilename, "r")
        else:
            f = open(infilename, "r")
        return f
    else:
        return None

def fill_database( infilename, name_cutoff=26, atom_count_cutoff=100):
    """compound is added if either name length of atom_count is below the
    corresponding cutoff value"""
    added = 0
    ignored = 0
    connection = sqlite.connect( Config.database_file)
    c = connection.cursor()
    f = _open_infile( infilename)
    if not f:
        raise ValueError( "File does not exist:", infilename)
    f.readline() # skip the first line
    i = 0
    for line in f:
        if i % 10000 == 0:
            print("done %8d, added %8d, ignored %8d" % (i,added,ignored))
        values = [x.strip() for x in line.strip().split("\t")]
        if len( values) != 4:
            print("Ignoring line:", line, file=sys.stderr, end='')
            continue
        cid, inchikey, smiles, name = [x.strip() for x in line.strip().split("\t")]
        c.execute( "DELETE FROM structures WHERE id=?", (cid,))
        if len( name) <= name_cutoff and len( [x for x in smiles if x.isupper()]) <= atom_count_cutoff and _allow_molecule( name, smiles):
            c.execute( "INSERT INTO structures (id,name,inchikey,smiles) VALUES (?,?,?,?);", (cid, name, inchikey, smiles))
            added += 1
        else:
            ignored += 1
        if i % 100 == 0:
            connection.commit()
        i += 1
    f.close()
    connection.commit()
    connection.close()
    return added, ignored


def get_compounds_from_database( database_file=None, **kw):
    """easy to use interface to the SQL - keyword arguments are converted to
    corresponding SQL commands.
    Examples:
    get_compounds_from_database( smiles='C1CCCCC1')
    get_compounds_from_database( inchi='1/C4H10/c1-3-4-2/h3-4H2,1-2H3')
    """
    for fname in (database_file, Config.database_file):
        if fname and os.path.exists(fname):
            break
    else:
        raise oasa_exceptions.oasa_error("Structure database not found. Try running 'python structure_database.py structures.txt.gz' in oasa directory to create the database file from default source.")

    if 'inchi' in kw:
        if not 'inchikey' in kw:
            from . import inchi_key
            kw['inchikey'] = inchi_key.key_from_inchi( kw['inchi'])
        del kw['inchi']
    search = ["%s=?" % k for k in kw.keys()]
    values = kw.values()
    tables = "structures"
    if "synonym" in kw:
        tables = "structures, synonyms"
        search.append( "structures.id=synonyms.id")
    if search:
        sql = "SELECT structures.* FROM %s WHERE %s" % (tables, " AND ".join( search))
    else:
        sql = "SELECT structures.* FROM %s" % (tables)
    connection = sqlite.connect( Config.database_file)
    c = connection.cursor()
    try:
        c.execute( sql, values)
    except sqlite.OperationalError as e:
        raise oasa_exceptions.oasa_error( "Error reading from structure database: '%s'" % e)
    ret = []
    c2 = connection.cursor()
    for row in c:
        c2.execute( "SELECT synonym FROM synonyms WHERE id=%d" % row[0])
        synonyms = list( c2)
        ret.append( row+([x[0] for x in synonyms],))
    c.close()
    c2.close()
    connection.close()
    return ret

def find_molecule_in_database( mol, database_file=None):
    """tries to find oasa.molecule mol in the database by using its InChiKey"""
    inchikey = inchi.generate_inchi_key( mol)[0]
    res = get_compounds_from_database( inchikey=inchikey)
    return res

def _allow_molecule( name, smile):
    if smile.count("-]") > 2:
        # more than 2 negative charges
        return False
    if smile.count("+]") > 2:
        # more than 2 positive charges
        return False
    if re.search( "\[\d", smile):
        return False
    if smile.count(".") > 2:
        return False
    disallowed = ["Tl","U","Ra","Rn","At","Po","Bi","Ta","W","Hf","La","Y","Sr","Nb","Zr","Eu","Lu","Ho","Ce","Nd","Dy","Th","Gd","Yb","Sm",'Es','Np','Pu','Pr','Md','Pm','Pa','Tb','Bk','Cm','Er','Cf','Fm','No',"Am","Ac","Tm","Lr","Fr"]
    for ch in disallowed:
        if ch in smile:
            return False
    # try smiles
##     try:
##         smiles.text_to_mol( smile, calc_coords=False)
##     except oasa_exceptions.oasa_invalid_atom_symbol, e:
##         return False
    return True

def filter_src_file( infilename, name_cutoff=26, atom_count_cutoff=1000):
    """print only those lines in infilename that would be allowed
    in structure_database using current settings"""
    added = 0
    ignored = 0
    f = _open_infile( infilename)
    if not f:
        raise ValueError
    f.readline() # skip the first line
    for line in f:
        values = [x.strip() for x in line.strip().split("\t")]
        if len( values) != 4:
            print("Ignoring line:", line, file=sys.stderr, end='')
            continue
        cid, inchikey, smiles, name = [x.strip() for x in line.strip().split("\t")]
        if len( name) <= name_cutoff and len( [x for x in smiles if x.isupper()]) <= atom_count_cutoff and _allow_molecule( name, smiles):
            print(line[:-1])
            added += 1
        else:
            ignored += 1
    f.close()
    return added, ignored

def _is_cid_in_db( cid, cursor):
    cursor.execute( "SELECT id FROM structures WHERE id=%s" % cid)
    if list( cursor):
        return True
    else:
        return False

def add_synonyms_old( fname, only_first=3):
    """this version checks each cid against databse, one by one,
    this is slow, but requires less memory"""
    f = _open_infile( fname)
    connection = sqlite.connect( Config.database_file)
    c = connection.cursor()
    c2 = connection.cursor()
    i = 0
    last_cid = None
    line_count = 0
    for line in f:
        line_count += 1
        parts = line.split()
        cid = parts[0]
        if last_cid != cid:
            last_cid == cid
            count = 1
            exists = _is_cid_in_db( cid, c2)
        else:
            count += 1
        if exists and count <= only_first:
            name = " ".join( parts[1:])
            c.execute( "INSERT INTO synonyms (id,name) VALUES (?,?);", (cid, name))
            i += 1
            if i % 500 == 0:
                connection.commit()
            if i % 10000 == 0:
                print("Added %d" % i)
        if line_count % 100000 == 0:
            print("-- %dk lines --" % (line_count // 1000))
        if line_count % 1000000 == 0:
            break
    connection.commit()
    return i


def add_synonyms( fname, only_first=3):
    """this version takes all cids into the memory for later comparison and is therefore
    much faster, it can however consume much memory for big databases"""
    f = _open_infile( fname)
    connection = sqlite.connect( Config.database_file)
    c = connection.cursor()
    c2 = connection.cursor()
    i = 0
    last_cid = None
    line_count = 0
    c.execute( "SELECT id FROM structures;")
    cids = set( [row[0] for row in c])
    for line in f:
        line_count += 1
        parts = line.split()
        cid = int( parts[0])
        if last_cid != cid:
            last_cid = cid
            count = 1
            exists = cid in cids
        else:
            count += 1
        if exists and count <= only_first:
            name = " ".join( parts[1:])
            c.execute( "INSERT INTO synonyms (id,synonym) VALUES (?,?);", (cid, name))
            i += 1
            if i % 500 == 0:
                connection.commit()
            if i % 10000 == 0:
                print("Added %d" % i)
        if line_count % 1000000 == 0:
            print("-- %dk lines --" % (line_count // 1000))
        #if line_count % 1000000 == 0:
        #    break
    connection.commit()
    return i



if __name__ == "__main__":
##     a,i = filter_src_file( 'selected_names_name50_inchi100.clean.txt')
##     print(a, i, file=sys.stderr)
##     sys.exit()

    from optparse import OptionParser
    op = OptionParser(usage="python %prog -c command [options] [structure file]")

    op.add_option( "-c", "--command", action="store",
                   dest="command", default="test",
                   help="command - what to do; one of 'test','update','rebuild','synonyms'")
    op.add_option( "-s", "--synonyms", action="store",
                   dest="synonyms", default="",
                   help="file containing synonyms")
    (options, args) = op.parse_args()

    if options.command == "test":
        print(get_compounds_from_database(inchi="1/C4H10/c1-3-4-2/h3-4H2,1-2H3"))
        print(get_compounds_from_database(smiles="O"))
        from . import smiles
        print(find_molecule_in_database(smiles.text_to_mol("C1CCC=CC1")))
        print(find_molecule_in_database(smiles.text_to_mol("c1ccccc1C")))
        print(get_compounds_from_database(synonym="toluene"))
    elif options.command in ('update','rebuild','synonyms'):
        if not options.command == 'synonyms':
            if len( args) >= 1:
                fname = args[0]
            else:
                print("You must supply a valid filename of a file containing structures to be read.", file=sys.stderr)
                sys.exit()
            if options.command == 'rebuild':
                if os.path.exists( Config.database_file):
                    try:
                        os.remove( Config.database_file)
                    except Exception as e:
                        print("Old database file could not be removed. Reason: %s" % e, file=sys.stderr)
                        print("Quitting..", file=sys.stderr)
                        sys.exit()
                create_database()
            added, ignored = fill_database( fname)
            print("Added: %d, Ignored: %d" % (added, ignored))
        if options.synonyms:
            print("Adding synonyms")
            #import profile
            #profile.run( 'add_synonyms( options.synonyms)')
            i = add_synonyms( options.synonyms)
            #i = 0
            print("Added %d synonyms" % i)
    else:
        if options.command not in ("test","rebuild","update"):
            print("Invalid action '%s'" % options.command, file=sys.argv)



#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

"""Minimalistic graph implementation.

Suitable for analysis of chemical problems.
"""

from __future__ import print_function

import copy
import operator
import warnings

from . import graph
from .diedge import diedge
from .vertex import vertex



class digraph(graph.graph):
  """Provide a minimalistic graph implementation.

  Suitable for analysis of chemical problems,
  even if some care was taken to make the graph work with nonsimple graphs,
  there are cases where it won't!
  """
  vertex_class = vertex
  edge_class = diedge


  def __str__( self):
    str = "digraph G(V,E), |V|=%d, |E|=%d" % ( len( self.vertices), len( self.edges))
    return str


  def add_edge( self, v1, v2, e=None):
    """adds an edge to a graph connecting vertices v1 and v2, if e argument is not given creates a new one.
    returns None if operation fails or the edge instance if successful"""
    i1 = self._get_vertex_index( v1)
    i2 = self._get_vertex_index( v2)
    if i1 is None or i2 is None:
      warnings.warn( "Adding edge to a vertex not present in graph failed (of course)", UserWarning, 3)
      return None
    # to get the vertices if v1 and v2 were indexes
    v1 = self.vertices[ i1]
    v2 = self.vertices[ i2]
    if not e:
      e = self.create_edge()
    e.set_vertices( (v1,v2))
    self.edges.add( e)
    v1.add_neighbor( v2, e)
    return e


  def get_diameter( self):
    diameter = 0
    best = None
    best_path = None
    for v in self.vertices:
      dist = self.mark_vertices_with_distance_from( v)
      if dist > diameter:
        diameter = dist
        best = v
        end = [x for x in self.vertices if 'd' in x.properties_ and x.properties_['d'] == dist][0]
        best_path = self.get_random_longest_path_numbered( v, end)

    print("path")
    best_path.reverse()
    for v in best_path:
      print(v)
    return diameter


  def get_connected_components( self):
    """returns the connected components of graph in a form o list of lists of vertices"""
    comp = set() # just processed component
    comps = []
    not_processed = set( self.vertices)
    if not_processed:
      recent = set() # [not_processed.pop()])
    processed = set()

    while not_processed:
      recent = set(j for i in [a.neighbors for a in recent]
                         for j in i) & not_processed
      processed = recent | comp
      for e in self.edges:
        v1, v2 = e.get_vertices()
        if (v1 in processed and v2 in not_processed):
          recent.add( v2)
        elif (v2 in processed and v1 in not_processed):
          recent.add( v1)
      if not recent:
        if comp:
          yield comp
        recent = set( [not_processed.pop()])
        comp = recent
      else:
        comp |= recent
        not_processed -= recent
    # when there is only one atom in the last piece it is not yielded in the loop
    yield comp


  def get_random_longest_path_numbered( self, start, end):
    """vertices have to be freshly marked with distance"""
    now = end
    path = []
    d = end.properties_['d']
    while now:
      d -= 1
      path.append( now)
      ns = [v for v in self.vertices if 'd' in v.properties_ and v.properties_['d'] == d and now in v.neighbors]
      if ns:
        now = ns[0]
      else:
        now = None
    return path


  def get_graphviz_text_dump( self):
    ret = '''digraph "dump" {
    fontpath="/usr/share/fonts/corefonts";
    ratio=compress
    nodesep=0.1;
    ranksep=0.3;
    rankdir=LR;
    node [shape=box,fontsize=10,fontname=Arial,height=0.3];
    '''
    for v in self.vertices:
      for n in v.neighbors:
        ret += '"%s" -> "%s";\n' % (v,n);
      ret += '"%s" [label="%s"];\n' % (v,v.value)
    ret += "}"
    return ret

